"""Interopbility with PyZX representation of ZX-calculus graphs."""

from tqec.interop.pyzx.plot import (
    draw_positioned_zx_graph_on as draw_positioned_zx_graph_on,
)
from tqec.interop.pyzx.plot import (
    plot_positioned_zx_graph as plot_positioned_zx_graph,
)
from tqec.interop.pyzx.plot import (
    pyzx_draw_positioned_zx_3d as pyzx_draw_positioned_zx_3d,
)
from tqec.interop.pyzx.positioned import PositionedZX as PositionedZX
from tqec.interop.pyzx.synthesis import (
    SynthesisStrategy as SynthesisStrategy,
)
from tqec.interop.pyzx.synthesis import (
    block_synthesis as block_synthesis,
)
from tqec.interop.pyzx.synthesis import (
    positioned_block_synthesis as positioned_block_synthesis,
)
from tqec.interop.pyzx.utils import (
    cube_kind_to_zx as cube_kind_to_zx,
)
from tqec.interop.pyzx.utils import (
    is_boundary as is_boundary,
)
from tqec.interop.pyzx.utils import (
    is_zx_no_phase as is_zx_no_phase,
)
from tqec.interop.pyzx.utils import (
    zx_to_basis as zx_to_basis,
)
from tqec.interop.pyzx.utils import (
    zx_to_pauli as zx_to_pauli,
)



from __future__ import annotations

import math
from dataclasses import dataclass, field
from typing import Any

import stim

from tqec.circuit.measurement_map import MeasurementRecordsMap
from tqec.circuit.qubit import GridQubit
from tqec.circuit.qubit_map import QubitMap
from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.compile.detectors.detector import Detector
from tqec.compile.observables.builder import Observable
from tqec.plaquette.rpng.rpng import PauliBasis
from tqec.utils.coordinates import StimCoordinates
from tqec.utils.exceptions import TQECError


@dataclass(frozen=True)
class DetectorAnnotation:
    """An annotation including all the necessary information to build a ``DETECTOR`` instruction."""

    coordinates: StimCoordinates
    measurement_offsets: list[int]

    def __post_init__(self) -> None:
        if any(m >= 0 for m in self.measurement_offsets):
            raise TQECError("Expected strictly negative measurement offsets.")

    def to_instruction(self) -> stim.CircuitInstruction:
        """Return the ``DETECTOR`` instruction represented by ``self``."""
        return stim.CircuitInstruction(
            "DETECTOR",
            [stim.target_rec(offset) for offset in self.measurement_offsets],
            self.coordinates.to_stim_coordinates(),
        )

    @staticmethod
    def from_detector(
        detector: Detector, measurement_records: MeasurementRecordsMap
    ) -> DetectorAnnotation:
        """Create a :class:`DetectorAnnotation` from a detector and measurement records."""
        return DetectorAnnotation(
            detector.coordinates,
            sorted([measurement_records[m.qubit][m.offset] for m in detector.measurements]),
        )


@dataclass(frozen=True)
class Polygon:
    """A polygon representing a stabilizer region in Crumble."""

    basis: PauliBasis
    qubits: frozenset[GridQubit]

    def _sorted_qubits(self) -> list[GridQubit]:  # pragma: no cover
        """Return the qubits in a sorted order that can be used to draw the polygon."""
        cx = sum(q.x for q in self.qubits) / len(self.qubits)
        cy = sum(q.y for q in self.qubits) / len(self.qubits)
        return sorted(self.qubits, key=lambda q: math.atan2(q.y - cy, q.x - cx))

    def to_crumble_url_string(self, qubit_map: QubitMap) -> str:
        """Convert the polygon to the representation in a crumble url."""
        # default grey color for polygons with no basis information
        rgba = [0, 0, 0, 0.25]
        rgba["xyz".index(self.basis.value)] = 1
        rgba_str = ",".join(str(i) for i in rgba)
        qubits_idx = [qubit_map[q] for q in self._sorted_qubits()]
        qubits_str = "_".join(str(i) for i in qubits_idx)
        return f"POLYGON({rgba_str}){qubits_str};"


@dataclass
class LayerNodeAnnotations:
    circuit: ScheduledCircuit | None = None
    detectors: list[DetectorAnnotation] = field(default_factory=list)
    observables: list[Observable] = field(default_factory=list)
    polygons: list[Polygon] = field(default_factory=list)

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of ``self``."""
        return {  # pragma: no cover
            "circuit_str": (str(self.circuit.get_circuit()) if self.circuit is not None else None),
            "detectors": self.detectors,
            "observables": self.observables,
            "polygons": self.polygons,
        }


@dataclass
class LayerTreeAnnotations:
    qubit_map: QubitMap | None = None

    @property
    def has_qubit_map(self) -> bool:  # pragma: no cover
        """Return ``True`` if the qubit map annotation has been set."""
        return self.qubit_map is not None

    def to_dict(self) -> dict[str, Any]:  # pragma: no cover
        """Return a dictionary representation of ``self``."""
        ret: dict[str, Any] = {"qubit_map": None}
        if self.qubit_map is not None:
            ret["qubit_map"] = {
                i: (
                    q.x,
                    q.y,
                )
                for i, q in self.qubit_map.i2q.items()
            }
        return ret



"""Defines templates representing logical qubits and its constituent parts."""

from collections.abc import Sequence

import numpy
import numpy.typing as npt
from typing_extensions import override

from tqec.templates.base import BorderIndices, RectangularTemplate
from tqec.templates.enums import TemplateBorder
from tqec.utils.exceptions import TQECError
from tqec.utils.scale import LinearFunction, PlaquetteScalable2D


class QubitTemplate(RectangularTemplate):
    """An error-corrected qubit.

    The below text represents this template for an input ``k == 2`` ::

        1  5  6  5  6  2
        7  9 10  9 10 11
        8 10  9 10  9 12
        7  9 10  9 10 11
        8 10  9 10  9 12
        3 13 14 13 14  4

    """

    @override
    def instantiate(
        self, k: int, plaquette_indices: Sequence[int] | None = None
    ) -> npt.NDArray[numpy.int_]:
        if plaquette_indices is None:
            plaquette_indices = list(range(1, self.expected_plaquettes_number + 1))

        ret = numpy.zeros(self.shape(k).to_numpy_shape(), dtype=numpy.int_)

        # The four corners
        ret[0, 0] = plaquette_indices[0]
        ret[0, -1] = plaquette_indices[1]
        ret[-1, 0] = plaquette_indices[2]
        ret[-1, -1] = plaquette_indices[3]
        # The up side
        ret[0, 1:-1:2] = plaquette_indices[4]
        ret[0, 2:-1:2] = plaquette_indices[5]
        # The left side
        ret[1:-1:2, 0] = plaquette_indices[6]
        ret[2:-1:2, 0] = plaquette_indices[7]
        # The center
        ret[1:-1:2, 1:-1:2] = plaquette_indices[8]
        ret[2:-1:2, 2:-1:2] = plaquette_indices[8]
        ret[1:-1:2, 2:-1:2] = plaquette_indices[9]
        ret[2:-1:2, 1:-1:2] = plaquette_indices[9]
        # The right side
        ret[1:-1:2, -1] = plaquette_indices[10]
        ret[2:-1:2, -1] = plaquette_indices[11]
        # The bottom side
        ret[-1, 1:-1:2] = plaquette_indices[12]
        ret[-1, 2:-1:2] = plaquette_indices[13]

        return ret

    @property
    @override
    def scalable_shape(self) -> PlaquetteScalable2D:
        return PlaquetteScalable2D(LinearFunction(2, 2), LinearFunction(2, 2))

    @property
    @override
    def expected_plaquettes_number(self) -> int:
        return 14

    @override
    def get_border_indices(self, border: TemplateBorder) -> BorderIndices:
        match border:
            case TemplateBorder.TOP:
                return BorderIndices(1, 5, 6, 2)
            case TemplateBorder.BOTTOM:
                return BorderIndices(3, 13, 14, 4)
            case TemplateBorder.LEFT:
                return BorderIndices(1, 7, 8, 3)
            case TemplateBorder.RIGHT:
                return BorderIndices(2, 11, 12, 4)


class QubitSpatialCubeTemplate(RectangularTemplate):
    """An error-corrected qubit that has all the spatial boundaries in the same basis.

    The below text represents this template for an input ``k == 4`` ::

         1   9  10   9  10   9  10   9  10   2
        11   5  17  13  17  13  17  13   6  21
        12  20  13  17  13  17  13  17  14  22
        11  16  20  13  17  13  17  14  18  21
        12  20  16  20  13  17  14  18  14  22
        11  16  20  16  19  15  18  14  18  21
        12  20  16  19  15  19  15  18  14  22
        11  16  19  15  19  15  19  15  18  21
        12   7  15  19  15  19  15  19   8  22
         3  23  24  23  24  23  24  23  24   4

    Warning:
        For ``k == 1``, this template does not include any of the plaquette
        that have an index in ``[13, 17]`` and so its instantiation has a "hole"
        in the plaquette indices.

    """

    @override
    def instantiate(
        self, k: int, plaquette_indices: Sequence[int] | None = None
    ) -> npt.NDArray[numpy.int_]:
        if plaquette_indices is None:
            plaquette_indices = list(range(1, self.expected_plaquettes_number + 1))

        shape = self.shape(k)
        ret = numpy.zeros(shape.to_numpy_shape(), dtype=numpy.int_)
        size = shape.x

        # The four corners
        ret[0, 0] = plaquette_indices[0]
        ret[0, -1] = plaquette_indices[1]
        ret[-1, 0] = plaquette_indices[2]
        ret[-1, -1] = plaquette_indices[3]
        # The up side
        ret[0, 1:-1:2] = plaquette_indices[8]
        ret[0, 2:-1:2] = plaquette_indices[9]
        # The left side
        ret[1:-1:2, 0] = plaquette_indices[10]
        ret[2:-1:2, 0] = plaquette_indices[11]
        # The center, which is the complex part.
        # Plaquette indices between 12 and 15 (both inclusive) are for plaquettes
        # on positions where (i + j) is even. Plaquettes indices between 16 and
        # 19 (both inclusive) are for plaquettes on positions where (i + j) is
        # odd. This is represented by an offset when (i + j) is odd.
        for i in range(1, size - 1):
            for j in range(1, size - 1):
                offset = 4 if (i + j) % 2 == 1 else 0
                # If the cell represented by (i, j) is:
                # - on the top (above the main diagonal and above the anti-diagonal)
                if i <= j and i <= (size - 1 - j):
                    ret[i, j] = plaquette_indices[12 + offset]
                # - on the right (above the main diagonal and below the anti-diagonal)
                elif i < j and i > (size - 1 - j):
                    ret[i, j] = plaquette_indices[13 + offset]
                # - on the bottom (below the main diagonal and below the anti-diagonal)
                elif i >= j and i >= (size - 1 - j):
                    ret[i, j] = plaquette_indices[14 + offset]
                # - on the left (below the main diagonal and above the anti-diagonal)
                elif i > j and i < (size - 1 - j):
                    ret[i, j] = plaquette_indices[15 + offset]

        ret[1, 1] = plaquette_indices[4]
        ret[1, -2] = plaquette_indices[5]
        ret[-2, 1] = plaquette_indices[6]
        ret[-2, -2] = plaquette_indices[7]
        # The right side
        ret[1:-1:2, -1] = plaquette_indices[20]
        ret[2:-1:2, -1] = plaquette_indices[21]
        # The bottom side
        ret[-1, 1:-1:2] = plaquette_indices[22]
        ret[-1, 2:-1:2] = plaquette_indices[23]

        return ret

    @property
    @override
    def scalable_shape(self) -> PlaquetteScalable2D:
        return PlaquetteScalable2D(LinearFunction(2, 2), LinearFunction(2, 2))

    @property
    @override
    def expected_plaquettes_number(self) -> int:
        return 24

    @override
    def get_border_indices(self, border: TemplateBorder) -> BorderIndices:
        match border:
            case TemplateBorder.TOP:
                return BorderIndices(1, 9, 10, 2)
            case TemplateBorder.BOTTOM:
                return BorderIndices(3, 23, 24, 4)
            case TemplateBorder.LEFT:
                return BorderIndices(1, 11, 12, 3)
            case TemplateBorder.RIGHT:
                return BorderIndices(2, 21, 22, 4)


class QubitVerticalBorders(RectangularTemplate):
    """Two vertical sides of neighbouring error-corrected qubits glued together.

    The below text represents this template for an input ``k == 2`` ::
        1 2
        5 7
        6 8
        5 7
        6 8
        3 4

    """

    @override
    def instantiate(
        self, k: int, plaquette_indices: Sequence[int] | None = None
    ) -> npt.NDArray[numpy.int_]:
        if plaquette_indices is None:
            plaquette_indices = list(range(1, self.expected_plaquettes_number + 1))
        ret = numpy.zeros(self.shape(k).to_numpy_shape(), dtype=numpy.int_)
        # The four corners
        ret[0, 0] = plaquette_indices[0]
        ret[0, -1] = plaquette_indices[1]
        ret[-1, 0] = plaquette_indices[2]
        ret[-1, -1] = plaquette_indices[3]
        # The left side
        ret[1:-1:2, 0] = plaquette_indices[4]
        ret[2:-1:2, 0] = plaquette_indices[5]
        # The right side
        ret[1:-1:2, -1] = plaquette_indices[6]
        ret[2:-1:2, -1] = plaquette_indices[7]
        return ret

    @property
    @override
    def scalable_shape(self) -> PlaquetteScalable2D:
        """Return a scalable version of the template shape."""
        return PlaquetteScalable2D(LinearFunction(0, 2), LinearFunction(2, 2))

    @property
    @override
    def expected_plaquettes_number(self) -> int:
        return 8

    @override
    def get_border_indices(self, border: TemplateBorder) -> BorderIndices:
        match border:
            case TemplateBorder.TOP | TemplateBorder.BOTTOM:
                raise TQECError(
                    f"Template {self.__class__.__name__} does not have "
                    f"repeating elements on the {border.name} border."
                )
            case TemplateBorder.LEFT:
                return BorderIndices(1, 5, 6, 3)
            case TemplateBorder.RIGHT:
                return BorderIndices(2, 7, 8, 4)


class QubitHorizontalBorders(RectangularTemplate):
    """Two horizontal sides of neighbouring error-corrected qubits glued together.

    The below text represents this template for an input ``k == 2`` ::
        1 5 6 5 6 2
        3 7 8 7 8 4

    """

    @override
    def instantiate(
        self, k: int, plaquette_indices: Sequence[int] | None = None
    ) -> npt.NDArray[numpy.int_]:
        if plaquette_indices is None:
            plaquette_indices = list(range(1, self.expected_plaquettes_number + 1))
        ret = numpy.zeros(self.shape(k).to_numpy_shape(), dtype=numpy.int_)
        # The four corners
        ret[0, 0] = plaquette_indices[0]
        ret[0, -1] = plaquette_indices[1]
        ret[-1, 0] = plaquette_indices[2]
        ret[-1, -1] = plaquette_indices[3]
        # The up side
        ret[0, 1:-1:2] = plaquette_indices[4]
        ret[0, 2:-1:2] = plaquette_indices[5]
        # The bottom side
        ret[-1, 1:-1:2] = plaquette_indices[6]
        ret[-1, 2:-1:2] = plaquette_indices[7]
        return ret

    @property
    @override
    def scalable_shape(self) -> PlaquetteScalable2D:
        return PlaquetteScalable2D(LinearFunction(2, 2), LinearFunction(0, 2))

    @property
    @override
    def expected_plaquettes_number(self) -> int:
        return 8

    @override
    def get_border_indices(self, border: TemplateBorder) -> BorderIndices:
        match border:
            case TemplateBorder.TOP:
                return BorderIndices(1, 5, 6, 2)
            case TemplateBorder.BOTTOM:
                return BorderIndices(3, 7, 8, 4)
            case TemplateBorder.LEFT | TemplateBorder.RIGHT:
                raise TQECError(
                    f"Template {self.__class__.__name__} does not have repeating "
                    f"elements on the {border.name} border."
                )



"""Defines a class analogous to ``cirq.Moment``.

This module defines :class:`Moment` that is very close to the external
`cirq.Moment <https://quantumai.google/reference/python/cirq/Moment>`_
class.

Internally, :class:`Moment` stores the instructions using ``stim.Circuit``
instead of using ``cirq`` data-structures.

"""

from __future__ import annotations

import re
from collections.abc import Callable, Iterable, Iterator, Sequence
from copy import deepcopy
from functools import cache
from typing import Any, cast

import stim

from tqec.circuit.qubit import count_qubit_accesses, get_used_qubit_indices
from tqec.utils.exceptions import TQECError
from tqec.utils.instructions import is_annotation_instruction


class MultipleOperationsOnSameQubitError(TQECError):
    def __init__(self, qubits: Sequence[int]):
        """Create a new instance of the exception.

        Args:
            qubits: qubit indices that are targeted by several instructions at the same timestep.

        """
        self._qubits = sorted(qubits)
        super().__init__(
            "Moment instances cannot be initialized with a stim.Circuit "
            "instance containing gates applied on the same qubit. Found "
            "multiple gates applied on the following qubits: "
            f"{self._qubits}."
        )

    @property
    def qubits(self) -> list[int]:
        """Return the qubits that are targeted by several instructions at the same timestep."""
        return self._qubits  # pragma: no cover


class Moment:
    """A collection of instructions that can be executed in parallel.

    This class is a collection of ``stim.CircuitInstruction`` instances that
    can all be executed in parallel. That means that it maintains the following
    invariant:

    For each instruction contained in any instance of this class, exactly
    one of the following assertions is true:

    1. The instruction is an annotation (e.g., ``QUBIT_COORDS``, ``DETECTOR``, ...),
    2. It is the only instruction of the :class:`Moment` instance to be applied
        on its targets. In other words, no other instructions in the :class:`Moment`
        instance can be applied on the targets this instruction is applied to.

    In practice, that means that this class match closely the definition of
    `cirq.Moment <https://quantumai.google/reference/python/cirq/Moment>`_.
    The only minor different is that ``cirq`` only uses the second assertion
    above (meaning that an annotation might push a quantum gate to the next
    moment, even though the annotation is never executed in hardware).

    """

    def __init__(
        self,
        circuit: stim.Circuit,
        used_qubits: set[int] | None = None,
        _avoid_checks: bool = False,
    ) -> None:
        """Initialize a :class:`Moment` instance.

        Args:
            circuit: collection of instructions representing the :class:`Moment`.
                It should represent a valid moment, see the class documentation
                for a detailed explanation of the pre-conditions.
            used_qubits: a set of qubit indices used in the provided ``circuit``.
                It is the user responsibility to ensure that this input is
                correct if it is provided, as any non-``None`` input will not be
                checked. Defaults to ``None``, which triggers a call to
                :func:`.get_used_qubit_indices` to initialise the indices.
            _avoid_checks: avoid checking the validity of the provided ``circuit``.
                This parameter can be used when the user knows that the provided
                ``circuit`` is a valid moment (and so checks the pre-conditions
                listed in the class documentation). Defaults to ``False``, which
                triggers a systematic check and might raise if the provided
                ``circuit`` is not a valid moment.

        Raises:
            TQECError: if the provided ``circuit`` contains one or more
                ``TICK`` instruction.
            TQECError: if the provided ``circuit`` contains at least 2
                non-annotation instructions that are applied on the same qubit
                target.
            TQECError: if the provided ``circuit`` contains a ``REPEAT``
                block instruction.

        """
        if not _avoid_checks:
            Moment.check_is_valid_moment(circuit)
        self._circuit: stim.Circuit = circuit
        self._used_qubits: set[int]
        if used_qubits is not None:
            self._used_qubits = used_qubits
        else:
            self._used_qubits = get_used_qubit_indices(self._circuit)

    @property
    def circuit(self) -> stim.Circuit:
        """Get the underlying circuit containing the operations."""
        return self._circuit

    @staticmethod
    def check_is_valid_moment(circuit: stim.Circuit) -> None:
        """Check if the provided circuit can be considered a valid moment.

        Args:
            circuit: instance to check.

        Raises:
            TQECError: if the provided ``circuit`` contains one or more
                ``TICK`` instruction.
            TQECError: if the provided ``circuit`` contains at least 2
                non-annotation instructions that are applied on the same qubit
                target.
            TQECError: if the provided ``circuit`` contains a ``REPEAT``
                block instruction.

        """
        if circuit.num_ticks > 0:
            raise TQECError(
                "Cannot initialize a Moment with a stim.Circuit instance "
                "containing at least one TICK instruction."
            )
        qubit_usage = count_qubit_accesses(circuit)
        multi_used_qubits = [qi for qi, usage_count in qubit_usage.items() if usage_count > 1]
        if multi_used_qubits:
            raise MultipleOperationsOnSameQubitError(multi_used_qubits)
        if any(isinstance(inst, stim.CircuitRepeatBlock) for inst in circuit):
            raise TQECError(
                "Moment instances should no contain any instance of stim.CircuitRepeatBlock."
            )

    @staticmethod
    def from_instructions(instructions: Iterable[stim.CircuitInstruction]) -> Moment:
        """Construct a :class:`Moment` instance from the provided ``instructions``."""
        circuit = stim.Circuit()
        for inst in instructions:
            circuit.append(inst)
        return Moment(circuit)

    @property
    def qubits_indices(self) -> set[int]:
        """Return the qubit indices this moment operates on.

        Note:
            Some instructions are considered annotations (e.g., ``QUBIT_COORDS``,
            see :data:`~tqec.circuit.qubit.NON_COMPUTATION_INSTRUCTIONS` for an
            exhaustive list). These instructions are ignored by this property,
            meaning that the qubits they operate on will only be returned by
            this property iff another non-annotation instruction is applied on
            said qubits.

        """
        return self._used_qubits

    def contains_instruction(self, instruction_name: str) -> bool:
        """Return ``True`` if ``self`` contains at least one operation with the provided name."""
        return any(instr.name == instruction_name for instr in self._circuit)

    def remove_all_instructions_inplace(self, instructions_to_remove: frozenset[str]) -> None:
        """Remove in-place all the instructions with a name in ``instructions_to_remove``."""
        new_circuit = stim.Circuit()
        for inst in self._circuit:
            if inst.name in instructions_to_remove:
                continue
            new_circuit.append(inst)
        self._circuit = new_circuit

    def __iadd__(self, other: Moment) -> Moment:
        """Add instructions in-place in ``self``."""
        both_sides_used_qubits = self._used_qubits.intersection(other._used_qubits)
        if both_sides_used_qubits:
            raise TQECError("Trying to add an overlapping quantum circuit to a Moment instance.")
        self._circuit += other._circuit
        return self

    def __add__(self, other: Moment) -> Moment:
        """Add instructions of ``self`` and ``other`` in a new instance."""
        both_sides_used_qubits = self._used_qubits.intersection(other._used_qubits)
        if both_sides_used_qubits:
            raise TQECError("Trying to add an overlapping quantum circuit to a Moment instance.")
        clone = deepcopy(self)
        clone += other
        return clone

    @staticmethod
    def _get_used_qubit_indices(
        targets: Iterable[int | stim.GateTarget],
    ) -> list[int]:
        qubits: list[int] = []
        for target in targets:
            if isinstance(target, int):
                qubits.append(target)
                continue
            # isinstance(target, stim.GateTarget)
            if target.is_qubit_target:
                assert isinstance(target.qubit_value, int)  # type checker is happy
                qubits.append(target.qubit_value)
        return qubits

    def append(
        self,
        name_or_instr: str | stim.CircuitInstruction,
        targets: Iterable[int | stim.GateTarget] | None = None,
        args: Iterable[float] | None = None,
    ) -> None:
        """Append an instruction to the :class:`Moment`.

        Note:
            if you append an annotation (e.g., ``DETECTOR`` or ``QUBIT_COORDS``)
            then you should use :meth:`append_annotation` that is more efficient.

        Args:
            name_or_instr: either the name of the instruction to append or the
                actual instruction. If the name is provided, ``targets`` and
                ``args`` are used to build the ``stim.CircuitInstruction``
                instance that will be appended. Else, they are not accessed.
            targets: if ``name_or_instr`` is a string representing the instruction
                name, this argument represent the targets the instruction should
                be applied on. Else, it is not used.
            args: if ``name_or_instr`` is a string representing the instruction
                name, this argument represent the arguments the instruction
                should be applied with. Else, it is not used.

        """
        if targets is None:
            targets = tuple()
        if args is None:
            args = tuple()

        instruction: stim.CircuitInstruction
        if isinstance(name_or_instr, str):
            instruction = stim.CircuitInstruction(name_or_instr, targets, args)
        else:
            instruction = name_or_instr

        if is_annotation_instruction(instruction):
            self.append_annotation(instruction)
            return

        # Checking Moment invariant
        instruction_qubits = Moment._get_used_qubit_indices(
            targets if isinstance(name_or_instr, str) else name_or_instr.targets_copy()
        )
        overlapping_qubits = self._used_qubits.intersection(instruction_qubits)
        if overlapping_qubits:
            raise TQECError(
                f"Cannot add {instruction} to the Moment due to qubit(s) "
                f"{overlapping_qubits} being already in use."
            )
        self._used_qubits.update(instruction_qubits)
        self._circuit.append(instruction)

    def append_annotation(self, annotation_instruction: stim.CircuitInstruction) -> None:
        """Append an annotation instruction to ``self``.

        This method is way more efficient than :meth:`append` to append an
        annotation. This is thanks to the fact that annotations are not using
        any qubit and so can be appended without checking that the instruction
        does not apply on already used qubits.

        Args:
            annotation_instruction: an annotation to append to the moment
                represented by ``self``.

        Raises:
            TQECError: if ``not is_annotation_instruction(annotation_instruction)``.

        """
        if not is_annotation_instruction(annotation_instruction):
            raise TQECError(
                "The method append_annotation only supports appending "
                f"annotations. Found instruction {annotation_instruction.name} "
                "That is not a valid annotation. Call append_instruction for "
                "generic instructions."
            )
        self._circuit.append(annotation_instruction)

    @property
    def instructions(self) -> Iterator[stim.CircuitInstruction]:
        """Iterator over all the instructions contained in ``self``."""
        # We can ignore the type error below because:
        # 1. if a Moment instance is created with a stim.CircuitRepeatBlock
        #    instance, it will raise an exception.
        # 2. `Moment.append` is the only method that may add an instance of
        #    stim.CircuitRepeatBlock, but its typing information explicitly
        #    prevents that case.
        # So we know for sure that there are only `stim.CircuitInstruction`
        # instances.
        yield from self._circuit  # type: ignore

    @property
    def num_measurements(self) -> int:
        """Return the number of measurements in the :class:`Moment` instance."""
        # Mypy is showing an error here:
        # error: Returning Any from function declared to return "int"
        # I do not understand why, but it probably has to do with Stim typing
        # so let's ignore it for the moment.
        return self._circuit.num_measurements  # type: ignore

    def filter_by_qubits(self, qubits_to_keep: Iterable[int]) -> Moment:
        """Return a new instance containing only the instructions applied on the provided qubits.

        Any operation from ``self`` that is applied to at least one qubit that is not in the
        provided ``qubits_to_keep`` will not be present in the returned :class:`Moment` instance.
        """
        qubits = frozenset(qubits_to_keep)
        used_qubits: set[int] = set()
        new_circuit = stim.Circuit()
        for instruction in self.instructions:
            targets: list[stim.GateTarget] = []
            for target_group in instruction.target_groups():
                qubit_targets = [
                    cast(int, t.qubit_value) for t in target_group if t.is_qubit_target
                ]
                if any(q not in qubits for q in qubit_targets):
                    continue
                targets.extend(target_group)
                used_qubits.update(qubit_targets)
            if targets:
                new_circuit.append(instruction.name, targets, instruction.gate_args_copy())
        return Moment(new_circuit, used_qubits=used_qubits, _avoid_checks=True)

    @property
    def is_empty(self) -> bool:
        """Return ``True`` if the :class:`Moment` instance is empty."""
        return len(self._circuit) == 0

    def __copy__(self) -> Moment:
        return Moment(  # pragma: no cover
            self._circuit,
            used_qubits=self._used_qubits,
            _avoid_checks=True,
        )

    def __deepcopy__(self, _: dict[Any, Any]) -> Moment:
        return Moment(
            self._circuit.copy(),
            used_qubits=self._used_qubits,
            _avoid_checks=True,
        )

    @staticmethod
    @cache
    def _qubit_target_regex() -> re.Pattern:
        # https://github.com/quantumlib/Stim/blob/main/doc/file_format_stim_circuit.md
        return re.compile(
            r"#.*"  # ignore comments
            r"|\[[^\]\r\n]*\]"  # ignore tags (can contain anything but ], \r, \n)
            r"|\([^)]*\)"  # ignore parens arguments
            # r"|\bREPEAT.*"  # Moment does not allow REPEAT blocks
            r"|\bMPAD.*"  # ignore MPAD instructions
            r"|rec\[-\d+\]"  # ignore measurement record targets
            r"|sweep\[\d+\]"  # ignore sweep bit targets
            r"|(\d+)",  # match remaining integers, i.e., qubit targets and Pauli targets
            # Note that tqec does not consider/support Pauli targets yet. Supporting them here to
            # be future-proof.
            re.IGNORECASE,
        )

    def with_mapped_qubit_indices(self, qubit_index_map: dict[int, int]) -> Moment:
        """Map the qubit **indices** on whom the :class:`Moment` instance is applied.

        Note:
            For performance, this method performs a regex matching to find and map qubit targets in
            the circuit string of ``self`` instead of iterating over the instructions.

        Args:
            qubit_index_map: the map used to modify the qubit targets.

        Returns:
            a modified copy of ``self`` with the qubit gate targets mapped according
            to the provided ``qubit_index_map``.

        """

        def replace_match(match):
            if (matched := match[1]) is not None:
                qubit = int(matched)
                return str(qubit_index_map.get(qubit, qubit))
            return match[0]

        return Moment(
            stim.Circuit(self._qubit_target_regex().sub(replace_match, str(self._circuit))),
            used_qubits={qubit_index_map[q] for q in self._used_qubits},
            _avoid_checks=True,
        )

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Moment) and self._circuit == other._circuit

    def __hash__(self) -> int:
        raise NotImplementedError(f"Cannot hash efficiently a {type(self).__name__}.")

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the :class:`Moment` instance.

        The dictionary is intended to be used as a JSON object.

        """
        return {
            "circuit": str(self._circuit),
            "used_qubits": list(self._used_qubits),
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> Moment:
        """Return a :class:`Moment` instance from its dictionary representation.

        Args:
            data: dictionary with the keys ``circuit`` and ``used_qubits``.

        Returns:
            a new instance of :class:`Moment` with the provided ``circuit`` and
            ``used_qubits``.

        """
        circuit = stim.Circuit(data["circuit"])
        used_qubits = set(data["used_qubits"])
        return Moment(circuit, used_qubits=used_qubits, _avoid_checks=True)


def iter_stim_circuit_without_repeat_by_moments(
    circuit: stim.Circuit, collected_before_use: bool = True
) -> Iterator[Moment]:
    """Iterate over the ``stim.Circuit`` by moments.

    A moment in a ``stim.Circuit`` is a sequence of instructions between two ``TICK``
    instructions. Note that ``stim.CircuitRepeatBlock`` instances are explicitly not
    supported and no such instance should appear in the provided circuit.

    Args:
        circuit: circuit to iterate over. Should not contain any ``REPEAT`` block.
        collected_before_use: if ``True``, the returned :class:`Moment` instances
            will contain a copy of the temporary ``stim.Circuit`` instance
            representing the moment. This is needed if the yielded
            :class:`Moment` instances are not used directly because the
            underlying ``stim.Circuit`` instance is cleared when resuming the
            generator.

    Yields:
        :class`Moment` instances.

    Raises:
        TQECError: if the provided ``circuit`` contains at least one
            ``stim.CircuitRepeatBlock`` instance.
        TQECError: if the provided ``circuit`` ``TICK`` instructions are not
            inserted such that instructions between two ``TICK`` instructions
            are always applied on disjoint sets of qubits.

    """
    copy_func: Callable[[stim.Circuit], stim.Circuit] = (
        (lambda c: c.copy()) if collected_before_use else (lambda c: c)
    )
    cur_moment = stim.Circuit()
    for inst in circuit:
        if isinstance(inst, stim.CircuitRepeatBlock):
            raise TQECError(
                "Found an instance of stim.CircuitRepeatBlock which is "
                "explicitly not supported by this method."
            )
        elif inst.name == "TICK":
            yield Moment(copy_func(cur_moment))
            cur_moment.clear()
        else:
            cur_moment.append(inst)
    # No need to copy the last moment
    yield Moment(cur_moment)



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING

from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.spatial import WithSpatialFootprint
from tqec.compile.blocks.layers.temporal import WithTemporalFootprint
from tqec.utils.scale import LinearFunction

if TYPE_CHECKING:
    from tqec.compile.blocks.layers.composed.sequenced import SequencedLayers


class BaseComposedLayer(WithSpatialFootprint, WithTemporalFootprint):
    """Base class representing a composed "layer".

    A composed layer is defined as a sequence (in time) of atomic layers. As
    such, composed layers are expected to have either a scalable time footprint
    (i.e., that grows with ``k``) or a constant time footprint that is strictly
    greater than ``1``.

    """

    @abstractmethod
    def all_layers(self, k: int) -> Iterable[BaseLayer]:
        """Return all the base layers represented by the instance.

        Returns:
            All the base layers represented by the instance. The returned
            iterable should have as many entries as ``self.timesteps(k)``.

        """
        pass

    @abstractmethod
    def to_sequenced_layer_with_schedule(
        self, schedule: tuple[LinearFunction, ...]
    ) -> SequencedLayers:
        """Split ``self`` into a :class:`.SequencedLayers` instance with the provided schedule.

        Args:
            schedule: duration of each of the layers in the returned :class:`.SequencedLayers`
                instance.

        Returns:
            an instance of :class:`.SequencedLayers` that is equivalent to ``self`` (same duration,
            same layers applied, ...) and that has the provided ``schedule``.

        """
        pass



from dataclasses import dataclass

from typing_extensions import override

from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.plaquette.compilation.passes.base import CompilationPass
from tqec.utils.exceptions import TQECError


@dataclass
class ScheduleMap:
    """Represent a map from schedules to schedules."""

    map: dict[int, int]

    def apply(self, circuit: ScheduledCircuit) -> ScheduledCircuit:
        """Apply the schedule map.

        This method will create a new
        :class:`~tqec.circuit.schedule.circuit.ScheduledCircuit` instance, only
        including the moments from the provided ``circuit`` if they are present
        as keys in ``self.map``, and mapping them to the schedule at the
        corresponding value.

        Args:
            circuit: circuit whose schedules should be modified.

        Returns:
            a new circuit with moments scheduled according to the values in
            ``self.map``.

        Raises:
            TQECError: if a moment in the provided ``circuit`` is scheduled
                at a time that is not present in ``self.map``.
            TQECError: if re-organising the schedules lead to an invalid
                circuit (i.e., at least one moment with overlapping operations).
                That can only happen if ``self.map.values()`` contains duplicate
                integers.

        """
        ret = ScheduledCircuit([], [], circuit.qubit_map)
        for moment_index, moment in circuit.scheduled_moments:
            if moment_index not in self.map:
                raise TQECError(
                    f"Found a moment scheduled at {moment_index} but this "
                    "schedule is not included in the schedule map."
                )
            new_schedule = self.map[moment_index]
            ret.add_to_schedule_index(new_schedule, moment)
        return ret


class ChangeSchedulePass(CompilationPass):
    def __init__(self, schedule_map: dict[int, int]):
        """Compilation pass changing the schedule of the provided quantum circuit."""
        super().__init__()
        self._map = ScheduleMap(schedule_map)

    @override
    def run(self, circuit: ScheduledCircuit, check_all_flows: bool = False) -> ScheduledCircuit:
        modified_circuit = self._map.apply(circuit)
        if check_all_flows:
            self.check_flows(circuit, modified_circuit)  # pragma: no cover
        return modified_circuit



"""Provides an implementation of various standard templates found when using topological quantum
error correction.
"""

from .fixed_boundary import (
    FixedBoundaryConventionGenerator as FixedBoundaryConventionGenerator,
)
from .fixed_bulk import FixedBulkConventionGenerator as FixedBulkConventionGenerator



"""Contains helpers to visualise a :class:`~tqec.compile.tree.tree.LayerTree` instance."""

from __future__ import annotations

import warnings
from collections.abc import Sequence
from dataclasses import dataclass

import stim
import svg
from typing_extensions import override

from tqec.circuit.qubit import GridQubit
from tqec.compile.blocks.layers.atomic.layout import LayoutLayer
from tqec.compile.observables.builder import Observable
from tqec.compile.tree.node import LayerNode, NodeWalker
from tqec.utils.exceptions import TQECError
from tqec.visualisation.computation.plaquette.grid import plaquette_grid_svg_viewer


@dataclass(frozen=True)
class VisualisationData:
    """Holds visualisation data for one layer.

    This dataclass holds data that will then be used to visualise one layer of the visualised
    :class:`~tqec.compile.tree.tree.LayerTree` instance.

    """

    layer: LayoutLayer
    start_moment: int
    end_moment: int
    observable: Observable | None = None

    def __post_init__(self) -> None:
        """Check that the instance is valid.

        Raises:
            AssertionError: when ``self.start_moment <= self.end_moment``.

        """
        assert self.start_moment <= self.end_moment

    def with_duration_offset(self, offset: int) -> VisualisationData:
        """Add a time offset to the time-related data in ``self``.

        This method is used to get a new instance of :class:`.VisualisationData` with a modified
        offset in time, for example when visualising each repetitions in a REPEAT loop.

        """
        return VisualisationData(
            self.layer,
            self.start_moment + offset,
            self.end_moment + offset,
            self.observable,
        )


class LayerVisualiser(NodeWalker):
    """A node walker that draws the visited :class:`~tqec.compile.tree.tree.LayerTree` instance."""

    def __init__(
        self,
        k: int,
        errors: Sequence[stim.ExplainedError] = tuple(),
        show_observable: int | None = None,
        font_size: float = 0.5,
        font_color: str = "red",
        top_left_qubit: GridQubit | None = None,
        bottom_right_qubit: GridQubit | None = None,
    ):
        """Create a :class:`.LayerVisualiser` instance.

        Args:
            k: scaling factor.
            errors: a (possibly empty) sequence of errors to draw on the resulting SVG
                representation.
            show_observable: also visualise the observable at the provided index if not None.
            font_size: size of the font used to write the moment range of each layer that is drawn.
            font_color: color of the font used to write the moment range of each layer that is
                drawn.
            top_left_qubit: qubit that should be at the top-left corner of the viewport. Can be used
                top only visualise part of a computation, or to add empty border. If not provided,
                the top-left qubit is automatically computed from the drawn computation.
            bottom_right_qubit: qubit that should be at the bottom-right corner of the viewport. Can
                be used top only visualise part of a computation, or to add empty border. If not
                provided, the bottom-right qubit is automatically computed from the drawn
                computation.

        """
        super().__init__()
        self._k = k
        self._stack: list[list[VisualisationData]] = [[]]
        self._errors: list[stim.ExplainedError] = list(errors)
        self._observable_index = show_observable
        self._font_size = font_size
        self._font_color = font_color
        self._top_left_qubit = top_left_qubit
        self._bottom_right_qubit = bottom_right_qubit

    @override
    def enter_node(self, node: LayerNode) -> None:
        if node.is_repeated:
            self._stack.append([])

    @override
    def exit_node(self, node: LayerNode) -> None:
        if not node.is_repeated:
            return
        if len(self._stack) < 2:
            raise TQECError(
                "Logical error: exiting a repeated node with less than 2 entries in the stack."
            )
        assert node.repetitions is not None
        repetitions = node.repetitions.integer_eval(self._k)
        elements = self._stack.pop()
        duration = elements[-1].end_moment - elements[0].start_moment
        # Repeat the body of the REPEAT loop.
        for r in range(repetitions):
            for e in elements:
                self._stack[-1].append(e.with_duration_offset(r * duration))

    @override
    def visit_node(self, node: LayerNode) -> None:
        if not node.is_leaf:
            return
        layer = node._layer
        assert isinstance(layer, LayoutLayer)
        start = self.current_moment
        end = start + layer.num_moments(self._k)
        observable: Observable | None = None
        if self._observable_index is not None:
            annotations = node._annotations.get(self._k)
            if annotations is not None:
                observable = next(
                    (
                        obs
                        for obs in node.get_annotations(self._k).observables
                        if obs.observable_index == self._observable_index
                    ),
                    None,
                )
        self._stack[-1].append(VisualisationData(layer, start, end, observable))

    @property
    def current_moment(self) -> int:
        """Return the index of the first moment on which something can be scheduled.

        Returns:
            1 + the index of the last used moment.

        """
        if self._stack[-1]:
            return self._stack[-1][-1].end_moment
        elif len(self._stack) > 1 and self._stack[-2]:
            return self._stack[-2][-1].end_moment
        else:
            return 0

    def get_moment_text(self, start: int, end: int) -> svg.Text:
        """Return an SVG representation of a text indicating the moments covered by [start, end].

        Args:
            start: initial moment.
            end: final (exclusive) moment.

        """
        return svg.Text(
            x=0.25,
            y=0.15,
            fill=self._font_color,
            font_size=self._font_size,
            text_anchor="start",
            dominant_baseline="hanging",
            text=f"Moments: {start} -> {end}",
        )

    def _get_errors_within(self, start: int, end: int) -> list[stim.ExplainedError]:
        return [
            err
            for err in self._errors
            if (start <= err.circuit_error_locations[0].tick_offset < end)
        ]

    @property
    def visualisations(self) -> list[str]:
        """Build and returns a list of SVG strings representing ``self``.

        Warning:
            Even though this is a property, non-trivial computations are made here. It is advised
            to only use that property once, when the visualisation is needed.

        """
        if len(self._stack) > 1:
            warnings.warn(
                "Trying to get the layer visualisations but the stack contains more than one "
                "element. You may get incorrect results. Did you forget to close a REPEAT block?"
            )
        if self._observable_index and not any(data.observable for data in self._stack[0]):
            raise TQECError(
                f"Observable index {self._observable_index} requested, but no observable "
                "with this index was found to be annotated in the layer visualisation data."
            )
        ret: list[str] = []
        for element in self._stack[0]:
            template, plaquettes = element.layer.to_template_and_plaquettes()
            instantiation = template.instantiate_list(self._k)
            drawers = plaquettes.collection.map_values(
                lambda plaq: plaq.debug_information.get_svg_drawer()
            )
            tlq, _ = element.layer.qubit_bounds

            svg_element = plaquette_grid_svg_viewer(
                instantiation,
                drawers,
                top_left_used_qubit=tlq.to_grid_qubit(self._k),
                view_box_top_left_qubit=self._top_left_qubit,
                view_box_bottom_right_qubit=self._bottom_right_qubit,
                errors=self._get_errors_within(element.start_moment, element.end_moment),
                observable=element.observable,
            )
            # Adding text to mark which TICKs are concerned.
            assert svg_element.elements is not None
            svg_element.elements.append(
                self.get_moment_text(element.start_moment, element.end_moment)
            )
            ret.append(svg_element.as_str())
        return ret



"""Contains exceptions related to :mod:`tqec.visualisation`."""

from tqec.utils.exceptions import TQECError


class TQECDrawingError(TQECError):
    """Exception raised when an error is encountered during drawing."""

    pass



"""Defines :class:`TopologicalComputationGraph` that represents a topological computation.

## Design choices

A few choices have been made during the design phase of this code, and they
deeply impact the public interface. This section explains what are those choices
and why they have been made.

### Substitution VS removal

Before the introduction of :mod:`tqec.compile.blocks`, pipes (or junctions) were
never strictly instantiated in any public-facing data-structure. Internally, the
pipe implementation was instantiated but was directly used to **replace**
plaquettes from the 2 blocks that were touched by the pipe. This replacement
was encoded as a "substitution" that was a mapping from (potentially negative)
layer indices to a mapping of plaquettes (i.e., a mapping from indices found in
template instantiation to :class:`~tqec.plaquette.plaquette.Plaquette` instances).

With the introduction of layers as the base data-structure to represent QEC
rounds, the substitution approach started to show weaknesses. For example when
the layer schedules of a cube and one of its spatial pipe do not match.

This situation arise with spatial junctions: the central cube implements a
schedule that is

1. initialisation layer,
2. repeat [memory layer],
3. measurement layer,

whereas some of the pipes need to implement

1. initialisation layer,
2. repeat [memory layer 1 alternated with memory layer 2],
3. measurement layer.

The fact that the repeated layer have a different body rules out any possibility
to substitute plaquettes.

To solve that issue, spatial pipes do not substitute plaquettes in cubes but
rather remove the cube boundary and implement its own layers.

For temporal pipes, the layers are replaced in-place within block instances.

"""

from pathlib import Path
from typing import Final

import stim

from tqec.compile.blocks.block import Block, merge_parallel_block_layers
from tqec.compile.blocks.enums import (
    SpatialBlockBorder,
    TemporalBlockBorder,
    border_from_signed_direction,
)
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.atomic.plaquettes import PlaquetteLayer
from tqec.compile.blocks.layers.composed.sequenced import SequencedLayers
from tqec.compile.blocks.positioning import (
    LayoutPipePosition2D,
    LayoutPosition2D,
    LayoutPosition3D,
)
from tqec.compile.detectors.database import DetectorDatabase
from tqec.compile.observables.abstract_observable import AbstractObservable
from tqec.compile.observables.builder import ObservableBuilder
from tqec.compile.tree.tree import LayerTree
from tqec.templates.enums import TemplateBorder
from tqec.utils.exceptions import TQECError
from tqec.utils.noise_model import NoiseModel
from tqec.utils.paths import DEFAULT_DETECTOR_DATABASE_PATH
from tqec.utils.position import BlockPosition3D, Direction3D, SignedDirection3D
from tqec.utils.scale import PhysicalQubitScalable2D


def substitute_plaquettes(
    target: PlaquetteLayer, source: PlaquetteLayer, source_border: TemplateBorder
) -> PlaquetteLayer:
    """Perform plaquette substitution on the provided layers.

    Note:
        Most of the time the ``target`` layer will represent a logical qubit and the ``source``
        layer will be a spatial junction template.

    Args:
        target: layer that will have one or more of its :class:`.Plaquette` overwritten.
        source: layer containing the :class:`.Plaquette` instances that will be copied over to
            ``target``.
        source_border: spatial border of the source that should be used. The opposite border will
            be used to update plaquettes on the target layer.

    Returns:
        a copy of ``target`` with plaquettes from ``source`` at its ``source_border.opposite()``
        border.

    """
    source_border_indices = source.template.get_border_indices(source_border)
    target_border_indices = target.template.get_border_indices(source_border.opposite())
    indices_mapping = source_border_indices.to(target_border_indices)
    plaquettes_mapping = {
        ti: target.plaquettes.collection[si]
        for si, ti in indices_mapping.items()
        if si in target.plaquettes.collection
    }
    new_plaquettes = target.plaquettes.with_updated_plaquettes(plaquettes_mapping)
    return PlaquetteLayer(target.template, new_plaquettes, target.trimmed_spatial_borders)


class TopologicalComputationGraph:
    def __init__(
        self,
        scalable_qubit_shape: PhysicalQubitScalable2D,
        observable_builder: ObservableBuilder,
        observables: list[AbstractObservable] | None = None,
    ) -> None:
        """Represent a topological computation with :class:`.Block` instances."""
        self._blocks: dict[LayoutPosition3D, Block] = {}
        # For fixed-bulk convention, temporal Hadamard pipe has its on space-time
        # extent. We need to keep track of the temporal pipes that are at the
        # same layer of at least one temporal Hadamard pipe.
        # We use the bottom cube position `z` to store the temporal pipe, s.t.
        # the pipe is actually at the position `z+0.5`
        self._temporal_pipes_at_hadamard_layer: dict[LayoutPosition3D, Block] = {}
        self._scalable_qubit_shape: Final[PhysicalQubitScalable2D] = scalable_qubit_shape
        self._observables: list[AbstractObservable] | None = observables
        self._observable_builder = observable_builder

    def add_cube(self, position: BlockPosition3D, block: Block) -> None:
        """Add a new cube at ``position`` implemented by the provided ``block``."""
        if not block.is_cube:
            raise TQECError(
                f"Cannot add the block as a cube. The provided block({block}) "
                "has at least one non-scalable dimension."
            )
        self._check_block_spatial_shape(block)
        layout_position = LayoutPosition3D.from_block_position(position)
        if layout_position in self._blocks:
            raise TQECError(
                "Cannot override a block with ``add_cube``. There is already "
                f"an entry at {layout_position}."
            )
        self._blocks[layout_position] = block

    def get_cube(self, position: BlockPosition3D) -> Block:
        """Recover the :class:`.Block` instance at the provided ``position``.

        Args:
            position: position of the block to recover.

        Raises:
            KeyError: if the provided ``position`` has no block associated.

        Returns:
            the :class:`.Block` instance at the provided ``position``.

        """
        layout_position = LayoutPosition3D.from_block_position(position)
        return self._blocks[layout_position]

    def _check_any_pipe(self, source: BlockPosition3D, sink: BlockPosition3D) -> None:
        """Check the validity of a pipe between ``source`` and ``sink``.

        Args:
            source: source of the pipe. Should be the "smallest" position.
            sink: destination of the pipe. Should be the "largest" position.

        Raises:
            TQECError: if ``source`` and ``sink`` are not neighbouring
                positions.
            TQECError: if ``not source < sink``.
            TQECError: if either ``source`` or ``sink`` has not been added
                to the graph.

        """
        if not source.is_neighbour(sink):
            raise TQECError(
                f"Trying to add a pipe between {source} and {sink} that are "
                "not neighbouring positions."
            )
        if not source < sink:
            raise TQECError(
                f"Trying to add a pipe between {source:=} and {sink:=} that "
                "are not correctly ordered. The following should be verified: "
                "source < sink."
            )
        source_layout_position = LayoutPosition3D.from_block_position(source)
        if source_layout_position not in self._blocks:
            raise TQECError(
                f"Cannot add a pipe between {source:=} and {sink:=}: the "
                "source is not in the graph."
            )
        sink_layout_position = LayoutPosition3D.from_block_position(sink)
        if sink_layout_position not in self._blocks:
            raise TQECError(
                f"Cannot add a pipe between {source:=} and {sink:=}: the sink is not in the graph."
            )

    def _check_spatial_pipe(self, source: BlockPosition3D, sink: BlockPosition3D) -> None:
        """Check the validity of a spatial pipe between ``source`` and ``sink``.

        Args:
            source: source of the pipe. Should be the "smallest" position.
            sink: destination of the pipe. Should be the "largest" position.

        Raises:
            TQECError: if ``source`` and ``sink`` are not neighbouring
                positions.
            TQECError: if ``not source < sink``.
            TQECError: if either ``source`` or ``sink`` has not been added
                to the graph.
            TQECError: if there is already a pipe between ``source`` and
                ``sink``.

        """
        self._check_any_pipe(source, sink)
        layout_position = LayoutPosition3D.from_pipe_position((source, sink))
        if layout_position in self._blocks:
            raise TQECError(
                "Cannot override a pipe with ``add_pipe``. "
                f"There is already a pipe at {layout_position}."
            )

    def _check_block_spatial_shape(self, block: Block) -> None:
        if block.scalable_shape != self._scalable_qubit_shape:
            raise TQECError(
                f"Expected a block shaped like a logical qubit "
                f"({self._scalable_qubit_shape}) but got {block.scalable_shape}."
            )

    def _trim_cube_spatial_borders(self, source: BlockPosition3D, sink: BlockPosition3D) -> None:
        """Trim the correct border from the cubes in ``source`` and ``sink``.

        This method trims 1 border on each of the cubes at the provided
        ``source`` and ``sink``.

        Args:
            source: source of the pipe. Should be the "smallest" position.
            sink: destination of the pipe. Should be the "largest" position.

        Raises:
            TQECError: if ``source`` and ``sink`` are not neighbouring
                positions.
            TQECError: if ``not source < sink``.
            TQECError: if either ``source`` or ``sink`` has not been added
                to the graph.
            TQECError: if there is already a pipe between ``source`` and
                ``sink``.

        """
        self._check_spatial_pipe(source, sink)
        juncdir = Direction3D.from_neighbouring_positions(source, sink)
        if juncdir not in Direction3D.spatial_directions():
            raise TQECError(
                f"The provided {source:=} and {sink:=} are not describing a "
                "valid spatial pipe. Spatial and temporal pipes should "
                "be handled separately."
            )
        psource = LayoutPosition3D.from_block_position(source)
        psink = LayoutPosition3D.from_block_position(sink)
        # Note that below the value of the ``toward_positive`` attribute of
        # SignedDirection3D is fixed by the condition that ``source < sink``.
        source_border = border_from_signed_direction(SignedDirection3D(juncdir, True))
        sink_border = border_from_signed_direction(SignedDirection3D(juncdir, False))
        assert isinstance(source_border, SpatialBlockBorder)
        assert isinstance(sink_border, SpatialBlockBorder)
        self._blocks[psource] = self._blocks[psource].with_spatial_borders_trimmed([source_border])
        self._blocks[psink] = self._blocks[psink].with_spatial_borders_trimmed([sink_border])

    def _substitute_part_of_spatial_pipe(
        self,
        pipe_pos: LayoutPosition3D[LayoutPipePosition2D],
        neighbouring_block_layer: PlaquetteLayer,
        spatial_block_border: SpatialBlockBorder,
        temporal_pipe_border: TemporalBlockBorder,
    ) -> None:
        """Substitute the plaquettes of the pipe at ``pipe_pos`` using ``neighbouring_block_layer``.

        The pipe in ``pipe_pos`` is modified in-line.

        Args:
            pipe_pos: position of the pipe that should be modified.
            neighbouring_block_layer: layer of the neighbouring block that
                should partially override the pipe.
            spatial_block_border: spatial border **of the block** that is being
                replaced by the pipe at ``pipe_pos``.
            temporal_pipe_border: temporal border of the pipe that should be
                partially replaced by ``neighbouring_block_layer``.

        Raises:
            KeyError: if ``pipe_pos not in self._blocks``.
            NotImplementError: if the pipe layer that should be partially
                substituted is not an instance of ``PlaquetteLayer``.

        """
        pipe_block = self._blocks[pipe_pos]
        pipe_layer_to_replace = pipe_block.get_temporal_layer_on_border(temporal_pipe_border)
        if not isinstance(pipe_layer_to_replace, PlaquetteLayer):
            raise NotImplementedError(
                "Due to the insertion of a temporal pipe, we need to replace "
                f"part of a pipe {temporal_pipe_border} border with part of the "
                f"temporal pipe. That is not possible because the "
                f"{temporal_pipe_border} spatial pipe border is not an instance "
                f"of {PlaquetteLayer.__name__}. Found an instance of "
                f"{type(pipe_layer_to_replace).__name__}."
            )
        # We now replace part of pipe_layer_to_replace
        replaced_pipe_layer = substitute_plaquettes(
            pipe_layer_to_replace,
            neighbouring_block_layer,
            spatial_block_border.to_template_border(),
        )
        replaced_block = pipe_block.with_temporal_borders_replaced(
            {temporal_pipe_border: replaced_pipe_layer}
        )
        assert replaced_block is not None, "No layer was removed"
        self._blocks[pipe_pos] = replaced_block

    def _replace_temporal_border(
        self,
        block_pos: BlockPosition3D,
        block_border: TemporalBlockBorder,
        layer: BaseLayer,
    ) -> None:
        pblock = LayoutPosition3D.from_block_position(block_pos)
        block = self._blocks[pblock]

        # First replace the layer on the temporal border of the block.
        layer_on_top_of_block = layer
        if block.trimmed_spatial_borders:
            layer_on_top_of_block = layer.with_spatial_borders_trimmed(
                block.trimmed_spatial_borders
            )
        new_block = block.with_temporal_borders_replaced({block_border: layer_on_top_of_block})
        assert new_block is not None, "No layer removal happened, only replacement"
        self._blocks[pblock] = new_block
        # Then, if the block has no trimmed spatial border (i.e., no spatial
        # pipes), we can return because the replacement is over.
        if not block.trimmed_spatial_borders:
            return
        # Else, we also need to replace part of the spatial pipe. Note that for
        # the moment this requires both the block and the spatial pipes to be
        # implemented using template / plaquettes.
        if not isinstance(layer, PlaquetteLayer):
            raise NotImplementedError(
                "Cannot substitute spatial pipe piece from a layer that is "
                f"not a {PlaquetteLayer.__name__} instance."
            )
        for trimmed_spatial_border in block.trimmed_spatial_borders:
            pipe_pos = LayoutPosition3D.from_block_and_signed_direction(
                block_pos, trimmed_spatial_border.value
            )
            self._substitute_part_of_spatial_pipe(
                pipe_pos, layer, trimmed_spatial_border, block_border
            )

    def _replace_temporal_borders(
        self, source: BlockPosition3D, sink: BlockPosition3D, block: Block
    ) -> None:
        self._check_any_pipe(source, sink)
        juncdir = Direction3D.from_neighbouring_positions(source, sink)
        if juncdir not in Direction3D.temporal_directions():
            raise TQECError(
                f"The provided {source:=} and {sink:=} are not describing a "
                "valid temporal pipe. Spatial and temporal pipes should "
                "be handled separately."
            )
        # Source
        self._replace_temporal_border(
            source,
            TemporalBlockBorder.Z_POSITIVE,
            block.get_atomic_temporal_border(TemporalBlockBorder.Z_NEGATIVE),
        )
        # Sink
        self._replace_temporal_border(
            sink,
            TemporalBlockBorder.Z_NEGATIVE,
            block.get_atomic_temporal_border(TemporalBlockBorder.Z_POSITIVE),
        )

    def add_pipe(self, source: BlockPosition3D, sink: BlockPosition3D, block: Block) -> None:
        """Add the provided block as a pipe between ``source`` and ``sink``.

        Raises:
            TQECError: if ``source`` and ``sink`` are not neighbouring
                positions.
            TQECError: if ``not source < sink``.
            TQECError: if there is already a pipe between ``source`` and
                ``sink``.
            TQECError: if ``block`` is not a valid pipe (i.e., has not
                exactly 2 scalable dimensions).

        """
        if not block.is_pipe:
            raise TQECError(
                "Cannot add as a pipe a block that is not a pipe. The provided "
                f"block ({block}) is not a pipe (i.e., does not have exactly 2 "
                "scalable dimensions)."
            )
        if block.is_temporal_pipe:
            self._check_block_spatial_shape(block)
            self._replace_temporal_borders(source, sink, block)
            block_trimmed_temporal_borders = block.with_temporal_borders_replaced(
                {
                    TemporalBlockBorder.Z_NEGATIVE: None,
                    TemporalBlockBorder.Z_POSITIVE: None,
                }
            )
            if block_trimmed_temporal_borders:
                u_pos = LayoutPosition3D.from_block_position(source)
                # We use the bottom cube position `z` to store the temporal pipe, s.t.
                # the pipe is actually at the position `z+0.5`
                self._temporal_pipes_at_hadamard_layer[u_pos] = block_trimmed_temporal_borders
        else:  # block is a spatial pipe
            self._trim_cube_spatial_borders(source, sink)
            key = LayoutPosition3D.from_pipe_position((source, sink))
            self._blocks[key] = block

    def to_layer_tree(self) -> LayerTree:
        """Merge layers happening in parallel at each time step.

        This method considers all the layers contained in added blocks (cubes and
        pipes) and merges them into a sequence of
        :class:`~tqec.compile.blocks.layers.atomic.layout.LayoutLayer` or
        :class:`~tqec.compile.blocks.layers.composed.base.BaseComposedLayer`
        wrapping :class:`~tqec.compile.blocks.layers.atomic.layout.LayoutLayer`
        instances.

        Returns:
            A tree representing the topological computation.

            The root node of the returned tree is an instance of
            :class:`~tqec.compile.blocks.layers.composed.sequenced.SequencedLayers`.
            Each child of the root node represents the computation happening during
            one block of time.

            Each child of the root node is also an instance of
            :class:`~tqec.compile.blocks.layers.composed.sequenced.SequencedLayers`.

        """
        zs = [pos.z for pos in self._blocks.keys()]
        min_z, max_z = min(zs), max(zs)
        blocks_by_z: list[dict[LayoutPosition2D, Block]] = [{} for _ in range(min_z, max_z + 1)]
        temporal_pipes_by_z: list[dict[LayoutPosition2D, Block]] = [
            {} for _ in range(min_z, max_z + 1)
        ]
        for pos, block in self._blocks.items():
            blocks_by_z[pos.z - min_z][pos.as_2d()] = block
        for pos, pipe in self._temporal_pipes_at_hadamard_layer.items():
            temporal_pipes_by_z[pos.z - min_z][pos.as_2d()] = pipe
        return LayerTree(
            SequencedLayers(
                [
                    SequencedLayers(
                        merge_parallel_block_layers(blocks, self._scalable_qubit_shape)
                        + merge_parallel_block_layers(pipes, self._scalable_qubit_shape),
                    )
                    for blocks, pipes in zip(blocks_by_z, temporal_pipes_by_z)
                ]
            ),
            abstract_observables=self._observables,
            observable_builder=self._observable_builder,
        )

    def generate_stim_circuit(
        self,
        k: int,
        noise_model: NoiseModel | None = None,
        manhattan_radius: int = 2,
        detector_database: DetectorDatabase | None = None,
        database_path: str | Path = DEFAULT_DETECTOR_DATABASE_PATH,
        do_not_use_database: bool = False,
        only_use_database: bool = False,
        reschedule_measurements: bool = True,
    ) -> stim.Circuit:
        """Generate the ``stim.Circuit`` from the compiled graph.

        Args:
            k: scale factor of the templates.
            noise_model: noise model to be applied to the circuit.
            manhattan_radius: radius considered to compute detectors.
                Detectors are not computed and added to the circuit if this
                argument is negative.
            detector_database: an instance to retrieve from / store in detectors
                that are computed as part of the circuit generation. If not given,
                the detectors are retrieved from/stored in the provided
                ``database_path``.
            database_path: specify where to save to after the calculation. This
                defaults to :data:`.DEFAULT_DETECTOR_DATABASE_PATH`
                if not specified. If detector_database is not passed in, the code
                attempts to retrieve the database from this location. The user
                may pass in the path either in str format, or as a Path instance.
            do_not_use_database: if ``True``, even the default database will not be used.
            only_use_database: if ``True``, only detectors from the database
                will be used. An error will be raised if a situation that is not
                registered in the database is encountered.
            reschedule_measurements: whether to reschedule measurements in a ``LayoutLayer``
                to be in the same moment. Since each plaquette may have its own measurement
                schedule, setting this may be necessary for hardware that requires
                measurements to be synchronous.

        Returns:
            A compiled stim circuit.

        """
        circuit = self.to_layer_tree().generate_circuit(
            k,
            manhattan_radius=manhattan_radius,
            detector_database=detector_database,
            database_path=database_path,
            do_not_use_database=do_not_use_database,
            only_use_database=only_use_database,
            reschedule_measurements=reschedule_measurements,
        )
        # If provided, apply the noise model.
        if noise_model is not None:
            circuit = noise_model.noisy_circuit(circuit)
        return circuit

    def generate_crumble_url(
        self,
        k: int,
        manhattan_radius: int = 2,
        detector_database: DetectorDatabase | None = None,
        add_polygons: bool = False,
    ) -> str:
        """Generate the Crumble URL from the compiled graph.

        Args:
            k: scaling factor.
            manhattan_radius: Parameter for the automatic computation of detectors.
                Should be large enough so that flows canceling each other to
                form a detector are strictly contained in plaquettes that are at
                most at a distance of ``manhattan_radius`` from the central
                plaquette. Detector computation runtime grows with this parameter,
                so you should try to keep it to its minimum. A value too low might
                produce invalid detectors.
            detector_database: existing database of detectors that is used to
                avoid computing detectors if the database already contains them.
                Default to `None` which result in not using any kind of database
                and unconditionally performing the detector computation.
            add_polygons: whether to include polygons in the Crumble URL. If
                ``True``, the polygons representing the stabilizers will be generated
                based on the RPNG information of underlying plaquettes and add
                to the Crumble URL.

        Returns:
            a string representing the Crumble URL of the quantum circuit.

        """
        return self.to_layer_tree().generate_crumble_url(  # pragma: no cover
            k, manhattan_radius, detector_database, add_polygons=add_polygons
        )



"""Defines functions to plot positioned ZX graphs and correlation surfaces on 3D axes."""

from __future__ import annotations

from dataclasses import astuple
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy
import numpy.typing as npt
from matplotlib.figure import Figure
from mpl_toolkits.mplot3d.axes3d import Axes3D
from pyzx import draw_3d
from pyzx.graph.graph_s import GraphS
from pyzx.pauliweb import PauliWeb

from tqec.interop.color import RGBA, TQECColor
from tqec.interop.pyzx.positioned import PositionedZX
from tqec.interop.pyzx.utils import is_boundary, is_hadamard, zx_to_pauli
from tqec.utils.enums import Pauli
from tqec.utils.position import Position3D

if TYPE_CHECKING:
    from tqec.computation.correlation import CorrelationSurface


def _node_color(g: GraphS, v: int) -> RGBA:  # pragma: no cover
    pauli = zx_to_pauli(g, v)
    assert pauli != Pauli.I
    return TQECColor(str(pauli)).rgba


def _positions_array(*positions: Position3D) -> npt.NDArray[numpy.int_]:  # pragma: no cover
    return numpy.array([astuple(p) for p in positions]).T


def draw_positioned_zx_graph_on(
    graph: PositionedZX,
    ax: Axes3D,
    *,
    node_size: int = 400,
    hadamard_size: int = 200,
    edge_width: int = 1,
) -> None:
    """Draw the :py:class:`~tqec.interop.pyzx.PositionedZX` on the provided axes.

    Args:
        graph: The positioned ZX graph to draw.
        ax: a 3-dimensional ax to draw on.
        node_size: The size of the node. Default is 400.
        hadamard_size: The size of the Hadamard transition. Default is 200.
        edge_width: The width of the edge. Default is 1.

    """
    g = graph.g
    pmap = graph.positions
    vis_nodes = [n for n in g.vertices() if not is_boundary(g, n)]
    vis_nodes_array = _positions_array(*[pmap[n] for n in vis_nodes])
    if vis_nodes_array.size > 0:
        ax.scatter(
            *vis_nodes_array,
            s=node_size,
            c=[_node_color(g, n).as_floats() for n in vis_nodes],
            alpha=1.0,
            edgecolors="black",
        )

    for edge in g.edges():
        pos_array = _positions_array(pmap[edge[0]], pmap[edge[1]])
        ax.plot(
            *pos_array,
            color="tab:gray",
            linewidth=edge_width,
        )
        if is_hadamard(g, edge):
            hadamard_position = numpy.mean(pos_array, axis=1)
            # use yellow square to indicate Hadamard transition
            ax.scatter(
                *hadamard_position,
                s=hadamard_size,
                c="yellow",
                alpha=1.0,
                edgecolors="black",
                marker="s",
            )
    ax.grid(False)
    for dim in (ax.xaxis, ax.yaxis, ax.zaxis):
        dim.set_ticks([])
    x_limits, y_limits, z_limits = ax.get_xlim3d(), ax.get_ylim3d(), ax.get_zlim3d()  # type: ignore

    plot_radius = 0.5 * max(abs(limits[1] - limits[0]) for limits in [x_limits, y_limits, z_limits])

    ax.set_xlim3d([numpy.mean(x_limits) - plot_radius, numpy.mean(x_limits) + plot_radius])
    ax.set_ylim3d([numpy.mean(y_limits) - plot_radius, numpy.mean(y_limits) + plot_radius])
    ax.set_zlim3d([numpy.mean(z_limits) - plot_radius, numpy.mean(z_limits) + plot_radius])


def draw_correlation_surface_on(
    correlation_surface: CorrelationSurface,
    graph: PositionedZX,
    ax: Axes3D,
    correlation_edge_width: int = 3,
) -> None:
    """Draw the correlation surface on the provided axes.

    Args:
        correlation_surface: The correlation surface to draw.
        graph: The positioned ZX graph to draw the correlation surface on.
        ax: The 3-dimensional ax to draw on.
        correlation_edge_width: The width of the correlation edges. Default is 3.

    """
    if correlation_surface.is_single_node:
        return

    pmap = graph.positions
    pauli_web = correlation_surface.to_pauli_web(graph.g)

    for edge, pauli in pauli_web.half_edges().items():
        up, vp = pmap[edge[0]], pmap[edge[1]]
        pos_array = _positions_array(up, vp)
        middle = numpy.mean(pos_array, axis=1).reshape(3, 1)
        start = _positions_array(up)
        ax.plot(
            *numpy.hstack([start, middle]),
            color=TQECColor(pauli).rgba.as_floats(),
            linewidth=correlation_edge_width,
        )


def plot_positioned_zx_graph(
    graph: PositionedZX,
    *,
    figsize: tuple[float, float] = (5, 6),
    title: str | None = None,
    node_size: int = 400,
    hadamard_size: int = 200,
    edge_width: int = 1,
) -> tuple[Figure, Axes3D]:
    """Plot the :py:class:`~tqec.interop.pyzx.positioned.PositionedZX` using matplotlib.

    Args:
        graph: The positioned ZX graph to plot.
        figsize: The figure size. Default is (5, 6).
        title: The title of the plot. Default to the name of the graph.
        node_size: The size of the node in the plot. Default is 400.
        hadamard_size: The size of the Hadamard square in the plot. Default is 200.
        edge_width: The width of the edge in the plot. Default is 1.

    Returns:
        A tuple of the figure and the axes.

    """
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111, projection="3d")

    draw_positioned_zx_graph_on(
        graph,
        ax,
        node_size=node_size,
        hadamard_size=hadamard_size,
        edge_width=edge_width,
    )

    if title:
        ax.set_title(title)
    fig.tight_layout()
    return fig, ax


def pyzx_draw_positioned_zx_3d(
    g: PositionedZX,
    id_labels: bool = True,
    pauli_web: PauliWeb[int, tuple[int, int]] | None = None,
) -> None:
    """Draw the positioned ZX graph in 3D with ``pyzx.draw_3d``.

    Args:
        g: The positioned ZX graph to draw.
        id_labels: Whether to show the vertex id labels. Default is True.
        pauli_web: The Pauli web to draw. Default is None.

    """
    plot_g = g.g.clone()
    for v in plot_g.vertices():
        position = g.positions[v]
        plot_g.set_qubit(v, position.x)
        plot_g.set_row(v, position.y)
        plot_g.set_vdata(v, "z", position.z)
    draw_3d(plot_g, labels=id_labels, pauli_web=pauli_web)



"""Defines functions and classes to generate and store sub-templates."""

from __future__ import annotations

import typing
from collections.abc import Sequence
from dataclasses import dataclass

import numpy
import numpy.typing as npt

from tqec.utils.exceptions import TQECError

SubTemplateType = npt.NDArray[numpy.int_]


@dataclass(frozen=True)
class UniqueSubTemplates:
    """Stores information on the sub-templates of a specific Manhattan radius present on a template.

    A sub-template is defined here as a portion of a
    :class:`~tqec.templates.base.Template` instantiation. In other words, a
    sub-template is a sub-array of the array resulting from calling the method
    :meth:`~tqec.templates.base.Template.instantiate` on any
    :class:`~tqec.templates.base.Template` instance. The size of this sub-array
    is defined by its `radius`, which is computed according to the Manhattan
    distance.

    For example, let's say you have the following array from calling
    :meth:`~tqec.templates.base.Template.instantiate` on a
    :class:`~tqec.templates.base.Template` instance: ::

        1  5  6  5  6  2
        7  9 10  9 10 11
        8 10  9 10  9 12
        7  9 10  9 10 11
        8 10  9 10  9 12
        3 13 14 13 14  4

    Focusing on the top-left ``9`` (coordinates ``(1, 1)`` in the array), the
    following sub-array is a sub-template of radius ``1``, centered on the
    ``(1, 1)`` coordinate of the above instantiation: ::

        1  5  6
        7  9 10
        8 10  9

    Still focusing on the top-left ``9``, the following sub-array is a
    sub-template of radius ``2``, centered on the ``(1, 1)`` coordinate of the
    above instantiation: ::

        .  .  .  .  .
        .  1  5  6  5
        .  7  9 10  9
        .  8 10  9 10
        .  7  9 10  9

    Note the inclusion of ``.`` that are here to represent ``0`` (i.e., the
    absence of plaquette at that particular point) because there is no
    plaquette in the original :class:`~tqec.templates.base.Template` instantiation.

    This dataclass efficiently stores all the sub-templates of a given
    `:class:`~tqec.templates.base.Template` instantiation and of a given ``radius``.

    Attributes:
        subtemplate_indices: an array that has the same shape as the
            original :class:`~tqec.templates.base.Template` instantiation but
            stores sub-template indices referencing sub-templates from the
            ``subtemplates`` attribute. The integers in this array do NOT represent
            plaquette indices.
        subtemplates: a store of sub-template (values) indexed by integers
            (keys) that link the sub-template center to the original
            template instantiation thanks to ``subtemplate_indices``.

    Raises:
        TQECError: if any index in ``self.subtemplate_indices`` is
            not present in ``self.subtemplates.keys()``.
        TQECError: if any of the sub-template shapes is non-square
            or of even width or length.
        TQECError: if not all the sub-template shapes in
            ``self.subtemplates.values()`` are equal.

    """

    subtemplate_indices: npt.NDArray[numpy.int_]
    subtemplates: dict[int, SubTemplateType]

    def __post_init__(self) -> None:
        # We do not need a valid subtemplate for the 0 index.
        indices = frozenset(numpy.unique(self.subtemplate_indices)) - {typing.cast(numpy.int_, 0)}
        if not indices.issubset(self.subtemplates.keys()):
            raise TQECError(
                "Found an index in subtemplate_indices that does not "
                "correspond to a valid subtemplate."
            )
        shape = next(iter(self.subtemplates.values())).shape
        if shape[0] != shape[1]:
            raise TQECError(
                "Subtemplate shapes are expected to be square. Found the "
                f"shape {shape} that is not a square."
            )
        if shape[0] % 2 == 0:
            raise TQECError(
                "Subtemplate shapes are expected to be squares with an odd "
                f"size length. Found size length {shape[0]}."
            )
        if not all(subtemplate.shape == shape for subtemplate in self.subtemplates.values()):
            raise TQECError(
                "All the subtemplates should have the exact same shape. "
                "Found one with a differing shape."
            )

    @property
    def manhattan_radius(self) -> int:
        """Return the Manhattan radius of the stored subtemplate."""
        any_subtemplate = next(iter(self.subtemplates.values()))
        height: int = any_subtemplate.shape[0]
        # height should be `2 * manhattan_radius + 1` so `height // 2` is
        # the Manhattan radius.
        return height // 2


def get_spatially_distinct_subtemplates(
    instantiation: npt.NDArray[numpy.int_],
    manhattan_radius: int = 1,
    avoid_zero_plaquettes: bool = True,
) -> UniqueSubTemplates:
    r"""Return a representation of all the distinct sub-templates of the provided Manhattan radius.

    Note:
        This function will likely be inefficient for large templates (i.e.,
        large values of :math:`k`) or for large Manhattan radiuses, both in
        terms of memory used and computation time.

        Right now, with

        - :math:`n` the width of the provided ``instantiation`` array,
        - :math:`m` the height of the provided ``instantiation`` array,
        - :math:`r` the provided Manhattan radius,

        it takes of the order of :math:`n m (2r+1)^2` memory and has to sort an
        array of :math:`nm` elements of size :math:`(2r+1)^2` in lexicographic
        order so require, in the worst case,
        :math:`O\left(nm\log(nm) (2r+1)^2\right)` runtime.

        Subclasses are invited to reimplement that method using a specialized
        algorithm (or hard-coded values) to speed things up.

        Some timings obtained on an AMD Ryzen 9 5950X: ::

            k = 10
            ----------------------------------------------------------------------
            radius   =         0 |         1 |         2 |         3 |         4
            time (s) =  0.000776 |  0.000992 |   0.00151 |   0.00227 |   0.00321
            ----------------------------------------------------------------------
            k = 20
            ----------------------------------------------------------------------
            radius   =         0 |         1 |         2 |         3 |         4
            time (s) =   0.00202 |   0.00356 |   0.00672 |    0.0116 |    0.0173
            ----------------------------------------------------------------------
            k = 40
            ----------------------------------------------------------------------
            radius   =         0 |         1 |         2 |         3 |         4
            time (s) =   0.00778 |    0.0156 |    0.0326 |    0.0574 |    0.0889
            ----------------------------------------------------------------------
            k = 80
            ----------------------------------------------------------------------
            radius   =         0 |         1 |         2 |         3 |         4
            time (s) =    0.0318 |    0.0706 |     0.147 |      0.27 |     0.426
            ----------------------------------------------------------------------
            k = 160
            ----------------------------------------------------------------------
            radius   =         0 |         1 |         2 |         3 |         4
            time (s) =     0.139 |     0.309 |     0.651 |      1.25 |      1.98
            ----------------------------------------------------------------------

    Args:
        instantiation: a 2-dimensional array representing the instantiated
            template on which sub-templates should be computed.
        manhattan_radius: radius of the considered ball using the Manhattan
            distance. Only squares with sides of ``2*manhattan_radius+1``
            plaquettes will be considered.
        avoid_zero_plaquettes: ``True`` if sub-templates with an empty plaquette
            (i.e., 0 value in the instantiation of the Template instance) at
            its center should be ignored. Default to ``True``.

    Returns:
        a representation of all the sub-templates found.

    """
    y, x = instantiation.shape
    extended_instantiation = numpy.pad(
        instantiation, manhattan_radius, "constant", constant_values=0
    )

    all_possible_subarrays: list[SubTemplateType] = []
    ignored_flattened_indices: list[int] = []
    considered_flattened_indices: list[int] = []
    for i in range(manhattan_radius, manhattan_radius + y):
        for j in range(manhattan_radius, manhattan_radius + x):
            # Do not generate anything if the center plaquette is 0 in the
            # original instantiation.
            if avoid_zero_plaquettes and extended_instantiation[i, j] == 0:
                ignored_flattened_indices.append(
                    (i - manhattan_radius) * x + (j - manhattan_radius)
                )
                continue
            considered_flattened_indices.append((i - manhattan_radius) * x + (j - manhattan_radius))
            all_possible_subarrays.append(
                extended_instantiation[
                    i - manhattan_radius : i + manhattan_radius + 1,
                    j - manhattan_radius : j + manhattan_radius + 1,
                ]
            )
    # Shape of the array provided to numpy.unique:
    #    (x * y, 2 * manhattan_radius + 1, 2 * manhattan_radius + 1)
    # Calling numpy.unique
    unique_situations, inverse_indices = numpy.unique(
        all_possible_subarrays, axis=0, return_inverse=True
    )

    # Note that the `inverse_indices` DO NOT include the ignored sub-templates because
    # their center was a 0 plaquette if `avoid_zero_plaquettes` is `True` so we
    # should reconstruct the full indices from `inverse_indices` and
    # `ignored_flattened_indices`.
    # By convention, the index 0 will represent the ignored sub-templates, so we
    # also have to shift the `inverse_indices` and `unique_situations` keys by 1.
    # Start by shifting by 1.
    inverse_indices += 1
    subtemplates_by_indices = {
        i + 1: typing.cast(npt.NDArray[numpy.int_], situation)
        for i, situation in enumerate(unique_situations)
    }
    final_indices: npt.NDArray[numpy.int_]
    if avoid_zero_plaquettes:
        final_indices = numpy.zeros((y * x,), dtype=numpy.int_)
        final_indices[ignored_flattened_indices] = 0
        final_indices[considered_flattened_indices] = inverse_indices
    else:
        final_indices = inverse_indices.astype(numpy.int_)
    return UniqueSubTemplates(final_indices.reshape((y, x)), subtemplates_by_indices)


@dataclass(frozen=True)
class Unique3DSubTemplates:
    """Stores the 3D sub-templates of a specific Manhattan radius in some templates.

    This class is a generalization of :class:`.UniqueSubTemplates` when several
    :class:`.Template` instances are concatenated in time. In that case,
    sub-templates of the last (in time) :class:`.Template` instance are not enough
    to represent the different "situations" that can be encountered.

    Attributes:
        subtemplate_indices: a 3-dimensional array with shape `(n, m, t)` where
            `(n, m)` is the shape of the original `Template` instantiation and
            `t` is the number of time-steps that are considered. This array
            stores **time independent** sub-template indices. The 3-dimensional
            sub-templates indices are all the entries on the third dimension,
            i.e., `t`-tuples indexed by `0<=i<n` and `0<=j<m`.
        subtemplates: a store of sub-template (values) indexed by `t`-tuples of
            integers (keys) that link the sub-template center to the original
            template instantiation thanks to `subtemplate_indices`.

    """

    subtemplate_indices: npt.NDArray[numpy.int_]
    subtemplates: dict[tuple[int, ...], SubTemplateType]

    def __post_init__(self) -> None:
        # Check that we have a 3-dimensional subtemplate_indices.
        shape = self.subtemplate_indices.shape
        if len(shape) != 3:
            raise TQECError(
                f"Expecting a 3-dimensional array but got the shape {shape} "
                f"that represents a {len(shape)}-dimensional array."
            )
        # Recover the number of timesteps and check that all the provided
        # subtemplates keys contain exactly the right number of entries.
        n, m, t = shape
        if any(len(k) != t for k in self.subtemplates.keys()):
            raise TQECError(
                f"Found {t} time slices in subtemplate_indices but got a key "
                f"in self.subtemplates that do not contain {t} entries."
            )
        # We do not need a valid subtemplate for the 0 index, but check that
        # all the other indices have their corresponding entry.
        zero = tuple(0 for _ in range(t))
        indices: frozenset[tuple[int, ...]] = frozenset(
            typing.cast(tuple[int, ...], tuple(arr))
            for arr in numpy.unique(self.subtemplate_indices.reshape(n * m, t), axis=0)
            if tuple(arr) != zero
        )
        if not indices.issubset(self.subtemplates.keys()):
            raise TQECError(
                "Found an index in subtemplate_indices that does not correspond "
                "to a valid subtemplate."
            )
        # Check that the provided subtemplate shapes are valid:
        # - square with odd sides in the spatial dimensions,
        # - 3-dimensional
        # - all equal
        shape = next(iter(self.subtemplates.values())).shape
        if shape[0] != shape[1]:
            raise TQECError(
                "Subtemplate shapes are expected to be square. Found the shape "
                f"{shape} that is not a square."
            )
        if shape[0] % 2 == 0:
            raise TQECError(
                "Subtemplate shapes are expected to be squares with an odd size "
                f"length. Found size length {shape[0]}."
            )
        if len(shape) != 3:
            raise TQECError(
                "Expecting 3-dimensional subtemplates but got a template with "
                f"{len(shape)} dimensions."
            )
        if not all(subtemplate.shape == shape for subtemplate in self.subtemplates.values()):
            raise TQECError(
                "All the subtemplates should have the exact same shape. "
                "Found one with a differing shape."
            )

    @property
    def manhattan_radius(self) -> int:
        """Return the Manhattan radius of the stored subtemplate."""
        any_subtemplate = next(iter(self.subtemplates.values()))
        height: int = any_subtemplate.shape[0]
        # height should be `2 * manhattan_radius + 1` so `height // 2` is
        # the Manhattan radius.
        return height // 2


def get_spatially_distinct_3d_subtemplates(
    instantiations: Sequence[npt.NDArray[numpy.int_]], manhattan_radius: int = 1
) -> Unique3DSubTemplates:
    r"""Return all the distinct 3-dimensional sub-templates of the provided Manhattan radius.

    Note:
        This function will likely be inefficient for large templates (i.e.,
        large values of :math:`k`) or for large Manhattan radiuses, both in
        terms of memory used and computation time.

        Right now, with

        - :math:`n` the width of the ``instantiations`` array entries,
        - :math:`m` the height of the ``instantiations`` array entries,
        - :math:`t` the number of time slices (``len(instantiations)``),
        - :math:`r` the provided Manhattan radius,

        it takes of the order of up to :math:`nmt(2r+1)^2` memory, has to sort
        :math:`t` arrays of :math:`nm` elements of size :math:`(2r+1)^2` in
        lexicographic order, so require, in the worst case,
        :math:`O\left(tnm\log(nm)(2r+1)^2\right)` runtime.

    Warning:
        This function assumes that the provided ``instantiations`` are compatible
        with each other. That means that they should all have the same shape and
        they should all be defined with respect to the same origin in order to
        ensure that stacking all the instantiations result in a 3-dimensional
        instantiation that is found in an actual QEC computation.

    Args:
        instantiations: a sequence of 2-dimensional arrays representing the
            instantiated templates on which sub-templates should be computed.
            They should all be "compatible" with each other. See warning in the
            function documentation.
        manhattan_radius: radius of the considered ball using the Manhattan
            distance. Only squares with sides of ``2*manhattan_radius+1``
            plaquettes will be considered.

    Returns:
        a representation of all the sub-templates found.

    """
    # Note: we explicitly do not avoid 0-indexed plaquette in the individual
    # 2-dimensional sub-templates (except the last one) because that led to
    # issues. The problem is that the computation is done independently for each
    # timeslice here, but we cannot "ignore" (i.e., return a 0-filled array) one
    # timeslice (again, except the last one) directly because the other
    # timeslices might need the potentially non-0 plaquettes around the center.
    # That issue arose when extending a qubit to perform a spatial junction with
    # stretched stabilizers (i.e., in fixed-boundary convention).
    unique_2d_subtemplates: list[UniqueSubTemplates] = [
        get_spatially_distinct_subtemplates(
            inst, manhattan_radius=manhattan_radius, avoid_zero_plaquettes=False
        )
        for i, inst in enumerate(instantiations)
    ]

    subtemplates_indices = numpy.stack(
        [u2ds.subtemplate_indices for u2ds in unique_2d_subtemplates], axis=2
    )
    n, m, t = subtemplates_indices.shape
    subtemplates: dict[tuple[int, ...], npt.NDArray[numpy.int_]] = {}

    unique_3d_indices = numpy.unique(subtemplates_indices.reshape(n * m, t), axis=0)
    # We might have 0 indices on some 2-dimensional slices. Because 0 will not be
    # a valid index for the 2-dimensional subtemplates we got from
    # get_spatially_distinct_subtemplates, pre-generate the corresponding array.
    zeros_2d = numpy.zeros((2 * manhattan_radius + 1, 2 * manhattan_radius + 1), dtype=numpy.int_)
    for indices in unique_3d_indices:
        if all(i == 0 for i in indices):
            continue
        # Get the list of subtemplates that should be stacked in the time dimension.
        subtemplates_2d_to_stack: list[npt.NDArray[numpy.int_]] = []
        for t, i in enumerate(indices):
            if i == 0:
                subtemplates_2d_to_stack.append(zeros_2d)
            else:
                subtemplates_2d_to_stack.append(unique_2d_subtemplates[t].subtemplates[i])
        # Stack the 2-dimensional subtemplates into a 3-dimensional one.
        subtemplate = numpy.stack(subtemplates_2d_to_stack, axis=2)
        # Add the 3-dimensional subtemplate to the mapping.
        subtemplates[tuple(indices.tolist())] = subtemplate
    return Unique3DSubTemplates(subtemplates_indices, subtemplates)



"""Defines :class:`Detector` to represent detectors."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import stim

from tqec.circuit.measurement import Measurement
from tqec.circuit.measurement_map import MeasurementRecordsMap
from tqec.utils.coordinates import StimCoordinates
from tqec.utils.exceptions import TQECError


@dataclass(frozen=True)
class Detector:
    """Represent a detector as a set of measurements and optional coordinates."""

    measurements: frozenset[Measurement]
    coordinates: StimCoordinates

    def __post_init__(self) -> None:
        if not self.measurements:
            raise TQECError("Trying to create a detector without any measurement.")

    def __hash__(self) -> int:
        return hash(self.measurements)

    def __eq__(self, rhs: object) -> bool:
        return (
            isinstance(rhs, Detector)
            and self.measurements == rhs.measurements
            and self.coordinates == rhs.coordinates
        )

    def __str__(self) -> str:
        measurements_str = "{" + ",".join(map(str, self.measurements)) + "}"
        return f"D{self.coordinates}{measurements_str}"

    def to_instruction(
        self, measurement_records_map: MeasurementRecordsMap
    ) -> stim.CircuitInstruction:
        """Return the ``stim.CircuitInstruction`` instance representing the detector in ``self``.

        Args:
            measurement_records_map: a map from qubits and qubit-local
                measurement offsets to global measurement offsets.

        Raises:
            TQECError: if any of the measurements stored in `self` is
                performed on a qubit that is not in the provided
                `measurement_records_map`.
            KeyError: if any of the qubit-local measurement offsets stored in
                `self` is not present in the provided `measurement_records_map`.

        Returns:
            the `DETECTOR` instruction representing `self`. Note that the
            instruction has the same validity region as the provided
            `measurement_records_map`.

        """
        measurement_records: list[stim.GateTarget] = []
        for measurement in self.measurements:
            if measurement.qubit not in measurement_records_map:
                raise TQECError(
                    f"Trying to get measurement record for {measurement.qubit} "
                    "but qubit is not in the measurement record map."
                )
            measurement_records.append(
                stim.target_rec(measurement_records_map[measurement.qubit][measurement.offset])
            )
        measurement_records.sort(key=lambda mr: mr.value, reverse=True)
        return stim.CircuitInstruction(
            "DETECTOR", measurement_records, self.coordinates.to_stim_coordinates()
        )

    def offset_spatially_by(self, x: int, y: int) -> Detector:
        """Offset the coordinates and all the qubits involved in `self`.

        Args:
            x: offset in the first spatial dimension.
            y: offset in the second spatial dimension.

        Returns:
            a new detector that has been spatially offset by the provided `x`
            and `y` offsets.

        """
        return Detector(
            frozenset(m.offset_spatially_by(x, y) for m in self.measurements),
            self.coordinates.offset_spatially_by(x, y),
        )

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the detector.

        Returns:
            a dictionary with the keys ``measurements`` and ``coordinates`` and
            their corresponding values.

        """
        return {
            "measurements": [m.to_dict() for m in self.measurements],
            "coordinates": self.coordinates.to_dict(),
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> Detector:
        """Return a detector from its dictionary representation.

        Args:
            data: dictionary with the keys ``measurements`` and ``coordinates``.

        Returns:
            a new instance of :class:`Detector` with the provided
            ``measurements`` and ``coordinates``.

        """
        measurements = frozenset(Measurement.from_dict(m) for m in data["measurements"])
        coordinates = StimCoordinates.from_dict(data["coordinates"])
        return Detector(measurements, coordinates)



from typing_extensions import override

from tqec.circuit.qubit import GridQubit
from tqec.compile.blocks.layers.atomic.layout import LayoutLayer
from tqec.compile.tree.annotations import Polygon
from tqec.compile.tree.node import LayerNode, NodeWalker
from tqec.plaquette.rpng.rpng import PauliBasis
from tqec.utils.array import to2dlist
from tqec.utils.position import Shift2D


class AnnotatePolygonOnLayerNode(NodeWalker):
    def __init__(self, k: int):
        """Node walker annotating potential polygons on each leaf node.

        Polygons are used when exporting a :class:`~tqec.compile.tree.tree.LayerTree` instance to a
        Crumble URL in order to have a visual hint representing the plaquettes directly in Crumble
        interface.

        Args:
            k: scaling factor.

        """
        self._k = k  # pragma: no cover

    @override
    def visit_node(self, node: LayerNode) -> None:
        if not node.is_leaf:  # pragma: no cover
            return
        assert isinstance(node._layer, LayoutLayer)
        node.get_annotations(self._k).polygons = generate_polygons_for_layout_layer(
            node._layer, self._k
        )


def generate_polygons_for_layout_layer(layer: LayoutLayer, k: int) -> list[Polygon]:
    """Generate the polygons that might be used to visualise stabilizers in Crumble."""
    template, plaquettes = layer.to_template_and_plaquettes()

    _indices = list(range(1, template.expected_plaquettes_number + 1))
    template_plaquettes = template.instantiate(k, _indices)
    increments = template.get_increments()

    polygons: list[Polygon] = []
    # The below line is not strictly needed, but makes type checkers happy with
    # type inference. See https://numpy.org/doc/stable/reference/typing.html#d-arrays
    # for more information on why this should be done.
    template_plaquettes_list: list[list[int]] = to2dlist(template_plaquettes)
    for row_index, line in enumerate(template_plaquettes_list):
        for column_index, plaquette_index in enumerate(line):
            if plaquette_index != 0:
                # Computing the offset that should be applied to each qubits.
                plaquette = plaquettes[plaquette_index]
                if plaquette.is_empty():
                    continue
                debug_info = plaquette.debug_information
                polygons_info = debug_info.get_polygons()
                if not polygons_info:
                    continue
                draw_polygons: dict[PauliBasis, list[GridQubit]]
                if isinstance(polygons_info, PauliBasis):
                    draw_polygons = {polygons_info: plaquette.qubits.data_qubits}
                else:
                    draw_polygons = polygons_info

                for basis, qubits in draw_polygons.items():
                    qubit_offset = Shift2D(
                        plaquette.origin.x + column_index * increments.x,
                        plaquette.origin.y + row_index * increments.y,
                    )
                    polygons.append(Polygon(basis, frozenset(q + qubit_offset for q in qubits)))

    # Shift the qubits of the returned scheduled circuit
    mincube, _ = layer.bounds
    eshape = layer.element_shape.to_shape_2d(k)
    # See: https://github.com/tqec/tqec/issues/525
    # This is a temporary fix to the above issue, we may need a utility function
    # to calculate shift to avoid similar issues in the future.
    shift = Shift2D(mincube.x * (eshape.x - 1), mincube.y * (eshape.y - 1))
    shifted_polygons = [Polygon(p.basis, frozenset(q + shift for q in p.qubits)) for p in polygons]

    return shifted_polygons



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, Mapping
from copy import deepcopy
from typing import Generic, TypeVar, cast

from typing_extensions import override

K = TypeVar("K")
V = TypeVar("V")
Vp = TypeVar("Vp")


class FrozenDefaultDict(Generic[K, V], Mapping[K, V]):
    def __init__(
        self,
        arg: Mapping[K, V] | Iterable[tuple[K, V]] | None = None,
        *,
        default_value: V | None = None,
    ) -> None:
        """Implement a ``defaultdict`` that cannot be mutated.

        This class re-defines all the mutating methods of `defaultdict` (i.e., all
        the mutating methods of `dict` and `__missing__`) in order to make any
        instance immutable.

        Note on re-defining `__missing__`:

        The standard `defaultdict` implementation is entirely based on the
        `__missing__` method (described here
        https://docs.python.org/3/library/collections.html#collections.defaultdict.__missing__)
        that is called when a user-provided key was not found in the defined keys.
        This `__missing__` method try to use `self.default_factory` to create a new
        value and inserts that new value in the dictionary. That last part is
        problematic for :class:`Plaquettes` and in particular to compare collections
        of :class:`Plaquettes` through `__hash__` and `__eq__`.

        Args:
            arg: a mapping from defined keys to values. Any key that is not present will be
                implicitly associated to the provided ``default_value``.
            default_value: value that will be returned (without copying) when a key not present in
                ``arg`` is queried.

        """
        super().__init__()
        self._dict: dict[K, V] = dict(arg) if arg is not None else dict()
        self._default_value = default_value

    def __missing__(self, key: K) -> V:
        if self._default_value is None:
            raise KeyError(key)
        return self._default_value

    @override
    def __getitem__(self, key: K) -> V:
        try:
            return self._dict[key]
        except KeyError:
            return self.__missing__(key)

    @override
    def __iter__(self) -> Iterator[K]:
        return iter(self._dict)

    @override
    def __len__(self) -> int:
        return len(self._dict)  # pragma: no cover

    @override
    def __contains__(self, key: object) -> bool:
        return self._dict.__contains__(key)

    def __or__(self, other: Mapping[K, V]) -> FrozenDefaultDict[K, V]:
        mapping = deepcopy(self._dict)
        mapping.update(other)
        return cast(
            FrozenDefaultDict[K, V],
            FrozenDefaultDict(mapping, default_value=self._default_value),
        )

    def __hash__(self) -> int:
        return hash(tuple(sorted(self.items())))  # pragma: no cover

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FrozenDefaultDict):
            return False  # pragma: no cover
        other = cast(FrozenDefaultDict[K, V], other)
        return (self._default_value == other._default_value) and self._dict == other._dict

    def has_default_value(self) -> bool:
        """Return ``True`` if ``self`` has a default value."""
        return self._default_value is not None  # pragma: no cover

    @property
    def default_value(self) -> V | None:
        """Get the default value returned when a queried key is missing."""
        return self._default_value

    def map_keys(self, callable: Callable[[K], K]) -> FrozenDefaultDict[K, V]:
        """Apply ``callable`` to each key and return a new instance with the modified keys."""
        return FrozenDefaultDict(
            {callable(k): v for k, v in self.items()},
            default_value=self._default_value,
        )

    def map_values(self, callable: Callable[[V], Vp]) -> FrozenDefaultDict[K, Vp]:
        """Apply ``callable`` to each value and return a new instance with the modified values."""
        default_value: Vp | None = None
        if self.default_value is not None:
            default_value = callable(self.default_value)
        defined_type: dict[K, Vp] = {k: callable(v) for k, v in self.items()}
        return cast(
            FrozenDefaultDict[K, Vp], FrozenDefaultDict(defined_type, default_value=default_value)
        )

    def map_keys_if_present(self, mapping: Mapping[K, K]) -> FrozenDefaultDict[K, V]:
        """Apply ``callable`` to each key and return a new instance with the modified keys."""
        return FrozenDefaultDict(  # pragma: no cover
            {mapping[k]: v for k, v in self.items() if k in mapping},
            default_value=self._default_value,
        )



from collections.abc import Sequence

import numpy
import numpy.typing as npt
from typing_extensions import override

from tqec.templates.base import BorderIndices, RectangularTemplate
from tqec.templates.enums import TemplateBorder
from tqec.utils.position import Shift2D
from tqec.utils.scale import LinearFunction, PlaquetteScalable2D


class FixedTemplate(RectangularTemplate):
    """A fixed template, only used internally for testing."""

    def __init__(
        self,
        indices: Sequence[Sequence[int]],
        default_increments: Shift2D | None = None,
    ) -> None:
        super().__init__(default_increments)
        self._indices: npt.NDArray[numpy.int_] = numpy.array([list(line) for line in indices])

    @override
    def instantiate(
        self, k: int = 0, plaquette_indices: Sequence[int] | None = None
    ) -> npt.NDArray[numpy.int_]:
        if plaquette_indices is None:
            plaquette_indices = list(range(1, self.expected_plaquettes_number + 1))

        return numpy.array(plaquette_indices)[self._indices]

    @property
    @override
    def scalable_shape(self) -> PlaquetteScalable2D:
        y, x = self._indices.shape
        return PlaquetteScalable2D(LinearFunction(0, x), LinearFunction(0, y))

    @property
    @override
    def expected_plaquettes_number(self) -> int:
        return max((max(line, default=0) for line in self._indices), default=0) + 1

    @override
    def get_border_indices(self, border: TemplateBorder) -> BorderIndices:
        raise NotImplementedError()



"""Defines code to automatically and efficiently compute detectors in a given time slice.

This module is crucial for ``tqec`` as it implements scalable functions to
automatically compute detectors in a provided time slice.

The implementation of this module is heavily based on sub-templates, that are
implemented in :mod:`tqec.templates.subtemplates`.

The main function defined by this module is
:func:`~.compute.compute_detectors_for_fixed_radius`. It allows to efficiently
compute all the detectors (represented as instances of
:class:`~.detector.Detector`) involving at least one of the measurements
contained in the last QEC round of the circuit that would be generated with
the provided :class:`~tqec.templates.base.Template` and
:class:`~tqec.plaquette.plaquette.Plaquette` instances.

Along with :func:`~.compute.compute_detectors_for_fixed_radius`, this module also
defines :class:`~.database.DetectorDatabase`. This is a class that is able to act
as a database for detectors and that can be used to:

- cache computations to avoid re-computing an already computed detector,
- ensure that the detectors in the final circuit are detectors from the provided
  database only (helps with reproducibility).

Implementation details can be found in the respective function/class
documentation.

"""

from .compute import (
    compute_detectors_for_fixed_radius as compute_detectors_for_fixed_radius,
)
from .database import DetectorDatabase as DetectorDatabase
from .detector import Detector as Detector



from __future__ import annotations

import argparse
from pathlib import Path

from typing_extensions import override

from tqec._cli.subcommands.base import TQECSubCommand
from tqec.computation.block_graph import BlockGraph


class CheckDaeTQECSubCommand(TQECSubCommand):
    @staticmethod
    @override
    def add_subcommand(
        main_parser: argparse._SubParsersAction[argparse.ArgumentParser],
    ) -> None:
        parser: argparse.ArgumentParser = main_parser.add_parser(
            "check_dae",
            description="Takes a .dae file in and check that it is correctly "
            "recognized by the tqec library.",
        )
        parser.add_argument("dae_file", help="The .dae file to check.", type=Path)
        parser.set_defaults(func=CheckDaeTQECSubCommand.execute)

    @staticmethod
    @override
    def execute(args: argparse.Namespace) -> None:
        dae_absolute_path: Path = args.dae_file.resolve()
        try:
            BlockGraph.from_dae_file(dae_absolute_path, graph_name=str(dae_absolute_path))
        except Exception as e:
            print("Failed to load the .dae file into a BlockGraph with the following error:")
            print(e)
        else:
            print("No issue found, the provided .dae file seems valid.")



from collections.abc import Callable
from typing import Any

import stim

from tqec.circuit.qubit import GridQubit
from tqec.circuit.qubit_map import QubitMap


def remap_qubit_indices(
    circuit: stim.Circuit,
    original_qubit_map: QubitMap | None = None,
    target_qubit_map: QubitMap | None = None,
    sort_key: Callable[[GridQubit], Any] = lambda q: (q.x, q.y),
) -> stim.Circuit:
    """Change qubit indices from ``circuit`` by using the provided ``target_qubit_map``.

    Args:
        circuit: original circuit that will be copied and changed by this
            function.
        original_qubit_map: qubit map representing the provided ``circuit``. If
            not provided, it will be automatically computed from ``circuit``.
        target_qubit_map: qubit map that should be used to index qubits in the
            circuit returned by this function. If not provided, this entry will
            be computed from ``original_qubit_map`` (or its default value) and
            ``sort_key`` by associating qubit indices from ``0`` to ``N-1`` to
            the ``N`` qubits in ``original_qubit_map`` in sorted order according
            to ``sort_key``.
        sort_key: callable used to sort qubits if ``target_qubit_map`` is not
            provided.

    Raises:
        KeyError: if ``original_qubit_map`` contains at least one qubit that is
            not in ``target_qubit_map``.

    Returns:
        a copy of the provided ``circuit`` with qubit indices changed to the
        provided ``target_qubit_map``.

    """
    if original_qubit_map is None:
        original_qubit_map = QubitMap.from_circuit(circuit)
    qubits = sorted(original_qubit_map.qubits, key=sort_key)
    if target_qubit_map is None:
        target_qubit_map = QubitMap.from_qubits(qubits)
    indices_mapping = {original_qubit_map[q]: target_qubit_map[q] for q in qubits}

    ret = stim.Circuit()
    for instruction in circuit:
        if isinstance(instruction, stim.CircuitRepeatBlock):
            ret.append(
                stim.CircuitRepeatBlock(
                    instruction.repeat_count,
                    remap_qubit_indices(
                        instruction.body_copy(),
                        original_qubit_map,
                        target_qubit_map,
                        sort_key,
                    ),
                )
            )
            continue
        ret.append(
            stim.CircuitInstruction(
                instruction.name,
                [
                    (indices_mapping[t.qubit_value] if t.qubit_value is not None else t)
                    for t in instruction.targets_copy()
                ],
                instruction.gate_args_copy(),
            )
        )
    return ret



from __future__ import annotations

from collections.abc import Generator
from enum import Enum, Flag, auto

from tqec.utils.exceptions import TQECError


class Orientation(Enum):
    """Either horizontal or vertical orientation."""

    HORIZONTAL = auto()
    VERTICAL = auto()

    def flip(self) -> Orientation:
        """Flip ``self``, returning the opposite orientation."""
        return Orientation.HORIZONTAL if self == Orientation.VERTICAL else Orientation.VERTICAL


class Basis(Enum):
    X = "X"
    Z = "Z"

    def flipped(self) -> Basis:
        """Flip ``self``, returning the opposite basis."""
        return Basis.X if self == Basis.Z else Basis.Z

    def to_pauli(self) -> Pauli:
        """Convert to the corresponding Pauli operator."""
        match self:
            case Basis.X:
                return Pauli.X
            case Basis.Z:
                return Pauli.Z

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self.value}"

    def __lt__(self, other: Basis) -> bool:
        return self.value < other.value


class PatchStyle(Enum):
    FixedBulk = auto()
    FixedBoundaryParity = auto()


# Flag is preferred over IntFlag for a number of reasons, including boundary control,
# i.e., preventing Pauli(4), etc. (enum.FlagBoundary is not available in Python 3.10)
class Pauli(Flag):
    """Pauli operators as bit flags of X and Z supports."""

    I = 0  # noqa: E741
    X = 1
    Z = 2
    Y = X | Z

    def flipped(self, condition: bool = True) -> Pauli:
        """Return the Pauli operator with X and Z supports flipped."""
        return ~self if condition else self

    def to_basis(self) -> Basis:
        """Convert to the corresponding `Basis`.

        Raises:
            TQECError: If the Pauli operator is not X or Z.

        Returns:
            The corresponding `Basis`.

        """
        match self:
            case Pauli.X:
                return Basis.X
            case Pauli.Z:
                return Basis.Z
            case _:
                raise TQECError(f"Cannot convert Pauli {self} to Basis.")

    def to_basis_set(self) -> set[Basis]:
        """Convert to the corresponding set of bases."""
        bases = list(Basis)
        return {bases[basis.value >> 1] for basis in self.iter_xz() if basis in self}

    @staticmethod
    def from_basis_set(bases: set[Basis]) -> Pauli:
        """Convert a set of bases to the corresponding Pauli operator."""
        return Pauli((Basis.X in bases) | ((Basis.Z in bases) << 1))

    def __invert__(self) -> Pauli:
        value = self.value
        return Pauli((value >> 1) | ((value % 2) << 1))

    def __str__(self) -> str:
        return "IXZY"[self.value]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self!s}"

    # Directly iterating over Pauli gives X, Z in Python 3.11+ but I, X, Y, Z in 3.10 due to a
    # behavior change in Flag. So we define these methods for consistent behavior across versions.
    @staticmethod
    def iter_xz() -> Generator[Pauli, None, None]:
        """Iterate over the X and Z Pauli operators."""
        yield from map(Pauli, range(1, 3))

    @staticmethod
    def iter_xzy() -> Generator[Pauli, None, None]:
        """Iterate over the X, Z, Y Pauli operators."""
        yield from map(Pauli, range(1, 4))

    @staticmethod
    def iter_ixzy() -> Generator[Pauli, None, None]:
        """Iterate over the I, X, Z, Y Pauli operators."""
        yield from map(Pauli, range(4))



"""Mapping from a positioned ZX graph to a block graph."""

from typing import cast

import pyzx as zx
from pyzx.utils import vertex_is_zx

from tqec.computation.block_graph import BlockGraph
from tqec.computation.cube import Cube, Port, YHalfCube, ZXCube
from tqec.computation.pipe import PipeKind
from tqec.interop.pyzx.positioned import PositionedZX
from tqec.interop.pyzx.utils import is_boundary, is_zx_no_phase
from tqec.utils.enums import Basis
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Direction3D


def positioned_block_synthesis(g: PositionedZX) -> BlockGraph:
    """Convert a positioned ZX graph to a :py:class:`~tqec.computation.block_graph.BlockGraph`.

    This strategy requires specifying the 3D positions of each vertex explicitly
    in the ZX graph. Then the conversion converts each vertex by looking at its
    nearest neighbors to infer the cube and pipe kind. The conversion maps each
    vertex to a cube in the block graph and each edge to a pipe connecting the
    corresponding cubes.

    The conversion process is as follows:

    1. Construct blocks for all the L/T/X-shape subgraphs in the ZX graph.
    2. Construct pipes connecting ports/Y to ports/Y nodes.
    3. Greedily construct the pipes until no more pipes can be inferred.
    4. If there are still nodes left, then choose orientation for an arbitrary node
       and repeat step 3 and 4 until all nodes are handled or conflicts are detected.

    Args:
        g: The positioned ZX graph to be converted to a block graph.

    Returns:
        The :py:class:`~tqec.computation.block_graph.BlockGraph` object
        converted from the ZX graph.

    Raises:
        TQECError: A valid block graph cannot be constructed.

    """
    nodes_to_handle = set(g.g.vertices())
    edges_to_handle = set(g.g.edges())

    bg = BlockGraph()
    # 1. Construct cubes for all the corner nodes in the ZX graph.
    _handle_corners(g, bg, nodes_to_handle)

    # 2. Construct pipes connecting ports/Y to ports/Y nodes.
    _handle_special_pipes(g, bg, nodes_to_handle, edges_to_handle)

    # 3. Greedily construct the pipes until no more pipes can be inferred.
    _greedily_construct_blocks(g, bg, nodes_to_handle, edges_to_handle)

    # 4. If there are still nodes left, then choose orientation for an arbitrary node
    # and repeat 3. Repeat 4 until all nodes are handled or conflicts are detected.
    _handle_leftover_nodes(g, bg, nodes_to_handle, edges_to_handle)

    # Sanity check for the block graph
    bg.validate()
    return bg


def _handle_corners(pg: PositionedZX, bg: BlockGraph, nodes_to_handle: set[int]) -> None:
    g = pg.g
    for v in g.vertices():
        directions = {pg.get_direction(u, v) for u in g.neighbors(v)}
        if len(directions) != 2:
            continue
        normal_direction = set(Direction3D.all_directions()).difference(directions).pop()
        normal_direction_basis = Basis.Z if g.type(v) == zx.VertexType.Z else Basis.X
        bases = [normal_direction_basis.flipped() for _ in range(3)]
        bases[normal_direction.value] = normal_direction_basis
        kind = ZXCube(*bases)
        bg.add_cube(pg[v], kind)
        nodes_to_handle.remove(v)


def _handle_special_pipes(
    pg: PositionedZX,
    bg: BlockGraph,
    nodes_to_handle: set[int],
    edges_to_handle: set[tuple[int, int]],
) -> None:
    g = pg.g

    for edge in set(edges_to_handle):
        u, v = edge
        if is_zx_no_phase(g, u) or is_zx_no_phase(g, v):
            continue
        cube_u = _port_or_y_cube(pg, u)
        cube_v = _port_or_y_cube(pg, v)
        bg.add_cube(cube_u.position, cube_u.kind, cube_u.label)
        bg.add_cube(cube_v.position, cube_v.kind, cube_v.label)
        pipe_kind = _choose_arbitrary_pipe_kind(pg, edge)
        bg.add_pipe(cube_u.position, cube_v.position, pipe_kind)
        nodes_to_handle.remove(u)
        nodes_to_handle.remove(v)
        edges_to_handle.remove(edge)


def _greedily_construct_blocks(
    pg: PositionedZX,
    bg: BlockGraph,
    nodes_to_handle: set[int],
    edges_to_handle: set[tuple[int, int]],
) -> None:
    num_nodes_left = len(nodes_to_handle) + 1
    while len(nodes_to_handle) < num_nodes_left:
        num_nodes_left = len(nodes_to_handle)
        _try_to_handle_edges(pg, bg, nodes_to_handle, edges_to_handle)


def _try_to_handle_edges(
    pg: PositionedZX,
    bg: BlockGraph,
    nodes_to_handle: set[int],
    edges_to_handle: set[tuple[int, int]],
) -> None:
    g = pg.g
    for edge in set(edges_to_handle):
        u, v = edge
        if u in nodes_to_handle and v in nodes_to_handle:
            continue
        ut, vt = g.type(u), g.type(v)
        can_infer_from_u = u not in nodes_to_handle and is_zx_no_phase(g, u)
        can_infer_from_v = v not in nodes_to_handle and is_zx_no_phase(g, v)
        if not can_infer_from_u and not can_infer_from_v:
            continue
        infer_from, other_node = (u, v) if can_infer_from_u else (v, u)
        ipos, opos = pg[infer_from], pg[other_node]
        cube_kind = cast(ZXCube, bg[ipos].kind)
        pipe_kind = PipeKind._from_cube_kind(
            cube_kind,
            pg.get_direction(u, v),
            can_infer_from_u,
            g.edge_type(edge) == zx.EdgeType.HADAMARD,
        )
        other_cube: Cube
        if is_zx_no_phase(g, other_node):
            other_cube_kind = _infer_cube_kind_from_pipe(
                pipe_kind, not can_infer_from_u, vt if can_infer_from_u else ut
            )
            # Check cube kind conflicts
            if other_node not in nodes_to_handle:
                existing_kind = bg[opos].kind
                if not other_cube_kind == existing_kind:
                    raise TQECError(f"Encounter conflicting cube kinds at {opos}: ")
            other_cube = Cube(opos, other_cube_kind)
        else:
            other_cube = _port_or_y_cube(pg, other_node)
        if other_node in nodes_to_handle:
            bg.add_cube(other_cube.position, other_cube.kind, other_cube.label)
            nodes_to_handle.remove(other_node)
        bg.add_pipe(ipos, opos, pipe_kind)
        edges_to_handle.remove(edge)


def _fix_kind_for_one_node(
    pg: PositionedZX,
    bg: BlockGraph,
    nodes_to_handle: set[int],
) -> None:
    g = pg.g
    sorted_nodes = sorted(nodes_to_handle, key=lambda n: pg[n])
    fix_node = next(n for n in sorted_nodes if is_zx_no_phase(g, n))
    fix_pos = pg[fix_node]
    fix_type = g.type(fix_node)
    # Special case: single node ZXGraph
    if g.vertex_degree(fix_node) == 0:
        specified_kind = (
            ZXCube.from_str("ZXZ") if fix_type == zx.VertexType.X else ZXCube.from_str("ZXX")
        )
    else:
        # the basis along the edge direction must be the opposite of the node kind
        basis = ["X", "Z"]
        neighbor = next(iter(g.neighbors(fix_node)))
        basis.insert(
            pg.get_direction(fix_node, neighbor).value,
            "X" if fix_type == zx.VertexType.Z else "Z",
        )
        specified_kind = ZXCube.from_str("".join(basis))
    bg.add_cube(fix_pos, specified_kind)
    nodes_to_handle.remove(fix_node)


def _handle_leftover_nodes(
    pg: PositionedZX,
    bg: BlockGraph,
    nodes_to_handle: set[int],
    edges_to_handle: set[tuple[int, int]],
) -> None:
    while nodes_to_handle:
        _fix_kind_for_one_node(pg, bg, nodes_to_handle)
        _greedily_construct_blocks(pg, bg, nodes_to_handle, edges_to_handle)


def _choose_arbitrary_pipe_kind(pg: PositionedZX, edge: tuple[int, int]) -> PipeKind:
    bases: list[str] = ["X", "Z"]
    direction = pg.get_direction(*edge)
    bases.insert(direction.value, "O")
    if pg.g.edge_type(edge) == zx.EdgeType.HADAMARD:
        bases.append("H")
    return PipeKind.from_str("".join(bases))


def _infer_cube_kind_from_pipe(
    pipe_kind: PipeKind,
    at_pipe_head: bool,
    vertex_type: zx.VertexType,
) -> ZXCube:
    """Infer the cube kinds from the pipe kind."""
    bases = [
        pipe_kind.get_basis_along(direction, at_pipe_head)
        for direction in Direction3D.all_directions()
    ]
    assert vertex_is_zx(vertex_type)
    bases[pipe_kind.direction.value] = Basis.Z if vertex_type == zx.VertexType.X else Basis.X
    return ZXCube(*cast(list[Basis], bases))


def _port_or_y_cube(pg: PositionedZX, v: int) -> Cube:
    g = pg.g
    if is_boundary(g, v):
        if v in g.inputs():
            label = f"IN_{g.inputs().index(v)}"
        elif v in g.outputs():
            label = f"OUT_{g.outputs().index(v)}"
        else:
            label = f"P_{v}"
        return Cube(pg[v], Port(), label)
    return Cube(pg[v], YHalfCube())



"""Read and write block graphs to and from Collada DAE files."""

from __future__ import annotations

import json
import pathlib
from collections.abc import Iterable
from dataclasses import dataclass
from typing import BinaryIO, cast

import collada
import collada.source
import numpy as np
import numpy.typing as npt

from tqec.computation.block_graph import BlockGraph, BlockKind, block_kind_from_str
from tqec.computation.correlation import CorrelationSurface
from tqec.computation.cube import CubeKind, Port, YHalfCube
from tqec.computation.pipe import PipeKind
from tqec.interop.collada._geometry import BlockGeometries, Face, get_correlation_surface_geometry
from tqec.interop.color import TQECColor
from tqec.utils.enums import Basis
from tqec.utils.exceptions import TQECError
from tqec.utils.position import FloatPosition3D, Position3D, SignedDirection3D
from tqec.utils.rotations import adjust_hadamards_direction, get_axes_directions, rotate_on_import
from tqec.utils.scale import round_or_fail

_ASSET_AUTHOR = "TQEC Community"
_ASSET_AUTHORING_TOOL_TQEC = "https://github.com/tqec/tqec"
_ASSET_UNIT_NAME = "inch"
_ASSET_UNIT_METER = 0.02539999969303608

_MATERIAL_SYMBOL = "MaterialSymbol"
_CORRELATION_SUFFIX = "_CORRELATION"


# SHARED FUNCTIONS
def _int_position_before_scale(pos: FloatPosition3D, pipe_length: float) -> Position3D:
    return Position3D(
        x=round_or_fail(pos.x / (1 + pipe_length), atol=0.35),
        y=round_or_fail(pos.y / (1 + pipe_length), atol=0.35),
        z=round_or_fail(pos.z / (1 + pipe_length), atol=0.35),
    )


def _offset_y_cube_position(pos: FloatPosition3D, pipe_length: float) -> FloatPosition3D:
    if np.isclose(pos.z - 0.5, np.floor(pos.z), atol=1e-9):
        pos = pos.shift_by(dz=-0.5)
    return FloatPosition3D(pos.x, pos.y, pos.z / (1 + pipe_length))


# DAE EXPORTER/IMPORTER
def read_block_graph_from_dae_file(
    filepath: str | pathlib.Path,
    graph_name: str = "",
) -> BlockGraph:
    """Read a Collada DAE file and construct a :class:`.BlockGraph` from it.

    Args:
        filepath: The input dae file path.
        graph_name: The name of the block graph. Default is an empty string.
        fix_shadowed_faces: Whether to fix the shadowed faces in the block graph. See
            :py:meth:`.BlockGraph.fix_shadowed_faces` for more details. Default is True.

    Returns:
        The constructed :py:class:`~tqec.computation.block_graph.BlockGraph` object.

    Raises:
        TQECError: If the COLLADA model cannot be parsed and converted to a block graph.

    """
    # Bring the mesh in
    mesh = collada.Collada(str(filepath))

    # Check some invariants about the DAE file
    if mesh.scene is None:
        raise TQECError("No scene found in the DAE file.")
    scene: collada.scene.Scene = mesh.scene

    if not (len(scene.nodes) == 1 and scene.nodes[0].name == "SketchUp"):
        raise TQECError(
            "The <visual_scene> node must have a single child node with the name 'SketchUp'."
        )

    sketchup_node: collada.scene.Node = scene.nodes[0]
    pipe_length: float | None = None
    parsed_cubes: list[tuple[FloatPosition3D, CubeKind, dict[str, int]]] = []
    parsed_pipes: list[tuple[FloatPosition3D, PipeKind, dict[str, int]]] = []

    # Handle nodes in scene/sketchup_node
    for node in sketchup_node.children:
        # If everything needed is present
        if (
            isinstance(node, collada.scene.Node)
            and node.matrix is not None
            and node.children is not None
            and len(node.children) == 1
            and isinstance(node.children[0], collada.scene.NodeNode)
        ):
            # Extract key info from collada scene
            instance = cast(collada.scene.NodeNode, node.children[0])
            library_node: collada.scene.Node = instance.node
            name: str = library_node.name

            # Skip the correlation surface nodes
            if name.endswith(_CORRELATION_SUFFIX):
                continue

            # Extract transformation and translation matrix
            transformation = _Transformation.from_4d_affine_matrix(node.matrix)
            translation = FloatPosition3D(*transformation.translation.tolist())
            axes_directions = get_axes_directions(transformation.rotation)
            kind = block_kind_from_str(name)

            # Rotations step 1. Skip if node's matrix not rotated
            # - If node's matrix YES rotated: check closer & make necessary adjustments
            if not np.allclose(transformation.rotation, np.eye(3), atol=1e-9):
                translation, kind = rotate_on_import(
                    transformation.rotation,
                    transformation.translation,
                    transformation.scale,
                    kind,
                )

            # Rotations step 2. Skip if hadamard points in positive direction
            if isinstance(kind, PipeKind):
                if axes_directions[str(kind.direction)] == -1 and "H" in str(kind):
                    kind = adjust_hadamards_direction(kind)

            # Direction, scaling and checks for pipes
            if isinstance(kind, PipeKind):
                # Get direction and scale of pipe
                pipe_direction = kind.direction
                scale = transformation.scale[pipe_direction.value]
                # Checks
                if pipe_length is None:
                    pipe_length = scale * 2.0
                elif not np.isclose(pipe_length, scale * 2.0, atol=1e-9):
                    raise TQECError("All pipes must have the same length.")
                expected_scale = np.ones(3)
                expected_scale[pipe_direction.value] = scale
                if not np.allclose(transformation.scale, expected_scale, atol=1e-9):
                    raise TQECError(
                        "Only the dimension along the pipe can be scaled, "
                        f"which is not the case at {translation}."
                    )
                # Append
                parsed_pipes.append((translation, kind, axes_directions))

            else:
                # Checks
                if not np.allclose(transformation.scale, np.ones(3), atol=1e-9):
                    raise TQECError(f"Cube at {translation} has a non-identity scale.")
                # Append
                parsed_cubes.append((translation, kind, axes_directions))

    pipe_length = 2.0 if pipe_length is None else pipe_length

    # Construct graph
    # Create graph
    graph = BlockGraph(graph_name)

    # Add cubes
    for pos, cube_kind, axes_directions in parsed_cubes:
        if isinstance(cube_kind, YHalfCube):
            graph.add_cube(
                _int_position_before_scale(_offset_y_cube_position(pos, pipe_length), pipe_length),
                cube_kind,
            )
        else:
            graph.add_cube(_int_position_before_scale(pos, pipe_length), cube_kind)
    port_index = 0

    # Add pipes
    for pos, pipe_kind, axes_directions in parsed_pipes:
        # Draw pipes in +1/-1 direction using position, kind of pipe, and directional pointers
        # from previous operations
        directional_multiplier = axes_directions[str(pipe_kind.direction)]
        head_pos = _int_position_before_scale(
            pos.shift_in_direction(pipe_kind.direction, -1 * directional_multiplier),
            pipe_length,
        )
        tail_pos = head_pos.shift_in_direction(pipe_kind.direction, 1 * directional_multiplier)

        # Add pipe
        if head_pos not in graph:
            graph.add_cube(head_pos, Port(), label=f"Port{port_index}")
            port_index += 1
        if tail_pos not in graph:
            graph.add_cube(tail_pos, Port(), label=f"Port{port_index}")
            port_index += 1
        graph.add_pipe(head_pos, tail_pos, pipe_kind)

    return graph


def write_block_graph_to_dae_file(
    block_graph: BlockGraph,
    file_like: str | pathlib.Path | BinaryIO,
    pipe_length: float = 2.0,
    pop_faces_at_directions: Iterable[SignedDirection3D | str] = (),
    show_correlation_surface: CorrelationSurface | None = None,
) -> None:
    """Write a :py:class:`~tqec.computation.block_graph.BlockGraph` to a Collada DAE file.

    Args:
        block_graph: The block graph to write to the DAE file.
        file_like: The output file path or file-like object that supports binary write.
        pipe_length: The length of the pipes in the COLLADA model. Default is 2.0.
        pop_faces_at_directions: Remove the faces at the given directions for all the blocks.
            This is useful for visualizing the internal structure of the blocks. Default is None.
        show_correlation_surface: The :py:class:`~tqec.computation.correlation.CorrelationSurface`
            to show in the block graph. Default is None.

    """
    directions: list[SignedDirection3D] = []
    for direction in pop_faces_at_directions:
        if isinstance(direction, str):
            directions.append(SignedDirection3D.from_string(direction))
        else:
            directions.append(direction)
    base = _BaseColladaData(directions)

    def scale_position(pos: Position3D) -> FloatPosition3D:
        return FloatPosition3D(*(p * (1 + pipe_length) for p in pos.as_tuple()))

    for cube in block_graph.cubes:
        if cube.is_port:
            continue

        scaled_position = scale_position(cube.position)
        if cube.is_y_cube and block_graph.has_pipe_between(
            cube.position, cube.position.shift_by(dz=1)
        ):
            scaled_position = scaled_position.shift_by(dz=0.5)

        matrix = np.eye(4, dtype=np.float32)
        matrix[:3, 3] = scaled_position.as_array()
        pop_directions = [
            SignedDirection3D(pipe.direction, cube == pipe.u)
            for pipe in block_graph.pipes_at(cube.position)
        ]
        base.add_block_instance(matrix, cube.kind, pop_directions)

    for pipe in block_graph.pipes:
        head_pos = scale_position(pipe.u.position)
        pipe_pos = head_pos.shift_in_direction(pipe.direction, 1.0)

        matrix = np.eye(4, dtype=np.float32)
        matrix[:3, 3] = pipe_pos.as_array()
        scales = [1.0, 1.0, 1.0]

        # We divide the scaling by 2.0 because the pipe's default length is 2.0.
        scales[pipe.direction.value] = pipe_length / 2.0
        matrix[:3, :3] = np.diag(scales)

        base.add_block_instance(matrix, pipe.kind)

    if show_correlation_surface is not None:
        base.add_correlation_surface(block_graph, show_correlation_surface, pipe_length)

    base.mesh.write(file_like)


# JSON IMPORTER
# Why no JSON exporter? No need. It can be done with blockgraph.to_dict
def read_block_graph_from_json(
    filepath: str | pathlib.Path,
    graph_name: str = "",
) -> BlockGraph:
    """Read a Collada JSON file and construct a :py:class:`.BlockGraph` from it.

    Args:
        filepath: The input dae file path.
        graph_name: The name of the block graph. Default is an empty string.

    Returns:
        The constructed :py:class:`.BlockGraph` object.

    Raises:
        TQECError: If the JSON file cannot be parsed and converted to a block graph.

    """
    # Read JSON file
    try:
        with open(filepath) as f:
            data = json.load(f)
    except Exception:
        raise TQECError("JSON file not found.")

    # Check JSON file has cubes and pipes
    try:
        cubes = data["cubes"]
        pipes = data["pipes"]
        if not len(cubes) > 0 or not len(pipes) > 0:
            raise TQECError("No cubes or pipes found.")
    except Exception:
        raise TQECError("JSON file is not appropriately formatted.")

    # Initialise list of cubes and pipes
    parsed_cubes: list[tuple[FloatPosition3D, CubeKind, dict[str, int]]] = []
    parsed_pipes: list[tuple[FloatPosition3D, FloatPosition3D, PipeKind, dict[str, int]]] = []

    # Get cubes data
    for cube in data["cubes"]:
        # Skip any "PORT" kind (cannot currently import them: error @ `block_kind_from_str` )
        if cube["kind"] == "PORT":
            continue

        # Enforce integers for position and transformation
        if not all([isinstance(i, int) for i in cube["position"]]):
            raise TQECError(
                f"Incorrect positioning for cube at {cube['position']}. All positional "
                "information must be composed of integer values."
            )

        if not all([isinstance(i, int) for row in cube["transform"] for i in row]):
            raise TQECError(
                f"Incorrect transformation matrix for cube at {cube['position']}. All "
                "elements in transformation matrix need to be integers."
            )

        # Get key spatial info
        translation = FloatPosition3D(*cube["position"])
        axes_directions = get_axes_directions(cube["transform"])
        kind = block_kind_from_str(cube["kind"])

        # Rotations step 1. Skip if node's matrix not rotated
        # - If node's matrix YES rotated: check closer & make necessary adjustments
        if not np.allclose(cube["transform"], np.eye(3), atol=1e-9):
            translation, kind = rotate_on_import(
                cube["transform"],
                cube["position"],
                np.array([1.0, 1.0, 1.0]),
                kind,
            )

        # Append to parsed cubes
        if isinstance(kind, CubeKind):
            parsed_cubes.append((translation, kind, axes_directions))

    # Get pipes data
    for pipe in data["pipes"]:
        # Enforce integers for position and transformation
        for pos in [pipe["u"], pipe["v"]]:
            if not all([isinstance(i, int) for i in pos]):
                raise TQECError(
                    f"Incorrect positioning for pipe at ({pipe['u']}). All positional "
                    "information must be composed of integer values."
                )

        if not all([isinstance(i, int) for row in pipe["transform"] for i in row]):
            raise TQECError(
                f"Incorrect transformation for pipe at ({pipe['u']}). All elements in "
                "transformation matrix need to be integers."
            )

        # Get key spatial info
        u_pos = FloatPosition3D(*pipe["u"])  # Equivalent to "translation" in cubes
        v_pos = FloatPosition3D(*pipe["v"])
        axes_directions = get_axes_directions(pipe["transform"])
        kind = block_kind_from_str(pipe["kind"])

        # Rotations step 1. Skip if node's matrix not rotated
        # - If node's matrix YES rotated: check closer & make necessary adjustments
        if not np.allclose(pipe["transform"], np.eye(3), atol=1e-9):
            u_pos, kind = rotate_on_import(
                pipe["transform"],
                pipe["position"],
                np.array([1.0, 1.0, 1.0]),
                kind,
            )

        # Rotations step 2. Skip if hadamard points in positive direction
        # Check kind is pipe
        if isinstance(kind, PipeKind):
            if axes_directions[str(kind.direction)] == -1 and "H" in str(kind):
                kind = adjust_hadamards_direction(kind)

        # Recheck kind since it might have been regenerated
        if isinstance(kind, PipeKind):
            parsed_pipes.append((u_pos, v_pos, kind, axes_directions))

    # Construct graph
    # Create graph
    graph = BlockGraph(graph_name)

    # Add cubes
    for pos, cube_kind, axes_directions in parsed_cubes:
        if isinstance(cube_kind, YHalfCube):
            graph.add_cube(
                _int_position_before_scale(_offset_y_cube_position(pos, 0.0), 0.0), cube_kind
            )
        else:
            graph.add_cube(_int_position_before_scale(pos, 0.0), cube_kind)
    port_index = 0

    # Add pipes
    for u_pos, v_pos, pipe_kind, axes_directions in parsed_pipes:
        # Write head_pos and tail_pos as Position3D
        head_pos = _int_position_before_scale(u_pos, 0)
        tail_pos = _int_position_before_scale(v_pos, 0)

        # Add pipe
        if head_pos not in graph:
            graph.add_cube(head_pos, Port(), label=f"Port{port_index}")
            port_index += 1
        if tail_pos not in graph:
            graph.add_cube(tail_pos, Port(), label=f"Port{port_index}")
            port_index += 1
        graph.add_pipe(head_pos, tail_pos, pipe_kind)

    return graph


# CLASSES
@dataclass(frozen=True)
class _BlockLibraryKey:
    """The key to access the library node in the Collada DAE file."""

    kind: BlockKind
    pop_faces_at_directions: frozenset[SignedDirection3D] = frozenset()

    def __str__(self) -> str:
        string = f"{self.kind}"
        if self.pop_faces_at_directions:
            string += "-without-"
            string += "-".join(str(d) for d in self.pop_faces_at_directions)
        return string


class _BaseColladaData:
    def __init__(
        self,
        pop_faces_at_directions: Iterable[SignedDirection3D] = (),
    ) -> None:
        """Encode the base model template.

        This class includes the definition of all the library nodes and the necessary material,
        geometry definitions.
        """
        self.mesh = collada.Collada()
        self.geometries = BlockGeometries()

        self.materials: dict[TQECColor, collada.material.Material] = {}
        self.geometry_nodes: dict[Face, collada.scene.GeometryNode] = {}
        self.root_node = collada.scene.Node("SketchUp", name="SketchUp")
        self.block_library: dict[_BlockLibraryKey, collada.scene.Node] = {}
        self.surface_library: dict[Basis, collada.scene.Node] = {}
        self._pop_faces_at_directions: frozenset[SignedDirection3D] = frozenset(
            pop_faces_at_directions
        )
        self._num_instances: int = 0

        self._create_scene()
        self._add_asset_info()
        self._add_materials()

    def _create_scene(self) -> None:
        scene = collada.scene.Scene("scene", [self.root_node])
        self.mesh.scenes.append(scene)
        self.mesh.scene = scene

    def _add_asset_info(self) -> None:
        if self.mesh.assetInfo is None:
            return
        self.mesh.assetInfo.contributors.append(
            collada.asset.Contributor(
                author=_ASSET_AUTHOR, authoring_tool=_ASSET_AUTHORING_TOOL_TQEC
            ),
        )
        self.mesh.assetInfo.unitmeter = _ASSET_UNIT_METER
        self.mesh.assetInfo.unitname = _ASSET_UNIT_NAME
        self.mesh.assetInfo.upaxis = collada.asset.UP_AXIS.Z_UP

    def _add_materials(self) -> None:
        """Add all the materials for different faces."""
        for face_color in TQECColor:
            rgba = face_color.rgba.as_floats()
            effect = collada.material.Effect(
                f"{face_color.value}_effect",
                [],
                "lambert",
                diffuse=rgba,
                emission=None,
                specular=None,
                transparent=rgba,
                transparency=rgba[3],
                ambient=None,
                reflective=None,
                double_sided=True,
            )
            self.mesh.effects.append(effect)

            material = collada.material.Material(
                f"{face_color.value}_material", f"{face_color.value}_material", effect
            )
            self.mesh.materials.append(material)
            self.materials[face_color] = material

    def _add_face_geometry_node(self, face: Face) -> collada.scene.GeometryNode:
        if face in self.geometry_nodes:
            return self.geometry_nodes[face]
        # Create geometry
        id_str = f"FaceID{len(self.geometry_nodes)}"
        positions = collada.source.FloatSource(
            id_str + "_positions", face.get_vertices(), ("X", "Y", "Z")
        )
        normals = collada.source.FloatSource(
            id_str + "_normals", face.get_normal_vectors(), ("X", "Y", "Z")
        )

        geom = collada.geometry.Geometry(self.mesh, id_str, id_str, [positions, normals])
        input_list = collada.source.InputList()
        input_list.addInput(0, "VERTEX", "#" + positions.id)
        input_list.addInput(1, "NORMAL", "#" + normals.id)
        triset = geom.createTriangleSet(Face.get_triangle_indices(), input_list, _MATERIAL_SYMBOL)
        geom.primitives.append(triset)
        self.mesh.geometries.append(geom)
        # Create geometry node
        inputs = [("UVSET0", "TEXCOORD", "0")]
        material = self.materials[face.color]
        geom_node = collada.scene.GeometryNode(
            geom, [collada.scene.MaterialNode(_MATERIAL_SYMBOL, material, inputs)]
        )
        self.geometry_nodes[face] = geom_node
        return geom_node

    def _add_block_library_node(
        self,
        block_kind: BlockKind,
        pop_faces_at_directions: Iterable[SignedDirection3D] = (),
    ) -> _BlockLibraryKey:
        pop_faces_at_directions = frozenset(pop_faces_at_directions) | self._pop_faces_at_directions
        key = _BlockLibraryKey(block_kind, pop_faces_at_directions)
        if key in self.block_library:
            return key
        faces = self.geometries.get_geometry(block_kind, pop_faces_at_directions)
        children = [self._add_face_geometry_node(face) for face in faces]
        key_str = str(key)
        node = collada.scene.Node(key_str, children, name=str(key.kind))
        self.mesh.nodes.append(node)
        self.block_library[key] = node
        return key

    def add_block_instance(
        self,
        transform_matrix: npt.NDArray[np.float32],
        block_kind: BlockKind,
        pop_faces_at_directions: Iterable[SignedDirection3D] = (),
    ) -> None:
        """Add an instance node to the root node."""
        key = self._add_block_library_node(block_kind, pop_faces_at_directions)
        child_node = collada.scene.Node(
            f"ID{self._num_instances}",
            name=f"instance_{self._num_instances}",
            transforms=[collada.scene.MatrixTransform(transform_matrix.flatten())],
        )
        point_to_node = self.block_library[key]
        instance_node = collada.scene.NodeNode(point_to_node)
        child_node.children.append(instance_node)
        self.root_node.children.append(child_node)
        self._num_instances += 1

    def _add_surface_library_node(self, basis: Basis) -> None:
        if basis in self.surface_library:
            return
        surface = get_correlation_surface_geometry(basis)
        geometry_node = self._add_face_geometry_node(surface)
        node = collada.scene.Node(surface.color.value, [geometry_node], name=surface.color.value)
        self.mesh.nodes.append(node)
        self.surface_library[basis] = node

    def add_correlation_surface(
        self,
        block_graph: BlockGraph,
        correlation_surface: CorrelationSurface,
        pipe_length: float = 2.0,
    ) -> None:
        # Needs to be imported here to avoid a circular import when importing tqec.interop
        from tqec.interop.collada._correlation import (  # noqa: PLC0415
            CorrelationSurfaceTransformationHelper,
        )

        helper = CorrelationSurfaceTransformationHelper(block_graph, pipe_length)

        for (
            basis,
            transformation,
        ) in helper.get_transformations_for_correlation_surface(correlation_surface):
            self._add_surface_library_node(basis)
            child_node = collada.scene.Node(
                f"ID{self._num_instances}",
                name=f"instance_{self._num_instances}_correlation_surface",
                transforms=[
                    collada.scene.MatrixTransform(transformation.to_4d_affine_matrix().flatten())
                ],
            )
            point_to_node = self.surface_library[basis]
            instance_node = collada.scene.NodeNode(point_to_node)
            child_node.children.append(instance_node)
            self.root_node.children.append(child_node)
            self._num_instances += 1


@dataclass(frozen=True)
class _Transformation:
    """Stores the translation, scale, rotation, and the composed affine matrix.

    For the reference of the transformation matrix, see https://en.wikipedia.org/wiki/Transformation_matrix.

    Attributes:
        translation: The length-3 translation vector.
        scale: The length-3 scaling vector, which is the scaling factor along each axis.
        rotation: The 3x3 rotation matrix.

    """

    translation: npt.NDArray[np.float32]
    scale: npt.NDArray[np.float32]
    rotation: npt.NDArray[np.float32]

    @staticmethod
    def from_4d_affine_matrix(mat: npt.NDArray[np.float32]) -> _Transformation:
        translation = mat[:3, 3]
        scale = np.linalg.norm(mat[:3, :3], axis=1)
        rotation = mat[:3, :3] / scale[None, :]
        return _Transformation(translation, scale, rotation)

    def to_4d_affine_matrix(self) -> npt.NDArray[np.float32]:
        mat = np.eye(4, dtype=np.float32)
        mat[:3, :3] = self.rotation * self.scale[None, :]
        mat[:3, 3] = self.translation
        return mat
